import os
import numpy
import graphicsPrimitives
import neuron
import numpy
cimport numpy
import itertools
import bisect
cimport cython


"""
The surfaces module
"""

cdef extern int find_triangles(double value0, double value1, double value2, double value3, double value4, double value5, double value6, double value7, double x0, double x1, double y0, double y1, double z0, double z1, double* out)
cdef extern double llgramarea(double* p0, double* p1, double* p2)
cdef extern double llpipedfromoriginvolume(double* p0, double* p1, double* p2)

cdef extern from "math.h":
    double sqrt(double)
    double fabs(double)



def _register_on_neighbor_map(the_map, pt, neighbor):
    # does not assume neighbor relations are bidirectional
    if pt in the_map:
        the_map[pt].append(neighbor)
    else:
        the_map[pt] = [neighbor]


max_chunks = 10000000

cdef int total_surface_tests = 0
cdef int corner_tests = 0

@cython.boundscheck(False)
@cython.wraparound(False)
#cdef bint contains_surface(int i, int j, int k, objdist, numpy.ndarray[numpy.float_t, ndim=1] xs, numpy.ndarray[numpy.float_t, ndim=1] ys, numpy.ndarray[numpy.float_t, ndim=1] zs, double dx, double r_inner, double r_outer, bint reject_if_outside, bint print_reject_reason=False):
def contains_surface(i, j, k, objdist, xs, ys, zs, dx, r_inner, r_outer, reject_if_outside, bint print_reject_reason=False):
    cdef bint has_neg = False
    cdef bint has_pos = False
    cdef double x, y, z

    global total_surface_tests, corner_tests
    total_surface_tests += 1

    # sphere tests added 2012-12-10
    cdef double xbar = xs[i] + dx / 2.
    cdef double ybar = ys[j] + dx / 2.
    cdef double zbar = zs[k] + dx / 2.

    d = fabs(objdist(xbar, ybar, zbar))
    if d <= r_inner: return True
    if d >= r_outer and reject_if_outside:
        if print_reject_reason:
            print 'would normally reject because at (%g, %g, %g): d = %g, r_outer = %g     (dx = %g)' % (xbar, ybar, zbar, d, r_outer, dx)
        else:
            return False



    # indeterminant from spheres; check corners
    corner_tests += 1
    #print 'corner_tests = %d out of total_surface_tests = %d' % (corner_tests, total_surface_tests)
    for x in xs[i : i + 2]:
        for y in ys[j : j + 2]:
            for z in zs[k : k + 2]:
                d = objdist(x, y, z)
                if print_reject_reason:
                    print 'at (%g, %g, %g): d = %g' % (x, y, z, d)
                if d <= 0:
                    has_neg = True
                if d >= 0:
                    has_pos = True
                if has_pos and has_neg:
                    return True

    return False



@cython.boundscheck(False)
@cython.wraparound(False)
#cdef int process_cell(int i, int j, int k, list objects, numpy.ndarray[numpy.float_t, ndim=1] xs, numpy.ndarray[numpy.float_t, ndim=1] ys, numpy.ndarray[numpy.float_t, ndim=1] zs, numpy.ndarray[numpy.float_t, ndim=1] tridata, int start, bint print_values=False):
def process_cell(int i, int j, int k, list objects, numpy.ndarray[numpy.float_t, ndim=1] xs, numpy.ndarray[numpy.float_t, ndim=1] ys, numpy.ndarray[numpy.float_t, ndim=1] zs, numpy.ndarray[numpy.float_t, ndim=1] tridata, int start, bint store_areas=False, numpy.ndarray[numpy.float_t, ndim=3] areas=None, bint print_values=False):
    cdef int new_index
    cdef double x, y, z, x1, y1, z1
    cdef list position
    x, y, z = xs[i], ys[j], zs[k]
    x1, y1, z1 = xs[i + 1], ys[j + 1], zs[k + 1]

    position = [(x, y, z),
                (x1, y, z),
                (x1, y1, z),
                (x, y1, z),
                (x, y, z1),
                (x1, y, z1),
                (x1, y1, z1),
                (x, y1, z1)]
    cdef double value0, value1, value2, value3, value4, value5, value6, value7
    value0, value1, value2, value3, value4, value5, value6, value7 = [min([objdist(*p) for objdist in objects]) for p in position]
    if print_values:
        print '(x, y, z) = (%7f, %7f, %7f); (x1, y1, z1) = (%7f, %7f, %7f)' % (x, y, z, x1, y1, z1)
        print '%7f %7f %7f %7f %7f %7f %7f %7f' % (value0, value1, value2, value3, value4, value5, value6, value7)
        print 'last obj distance to position[4]: ', objects[len(objects)-1](*position[4])
        print 'distance to position[4] with everything but the last object:', min([objdist(*position[4]) for objdist in objects[:len(objects)-1]])
        print 'distance to position[4] with everything:', min([objdist(*position[4]) for objdist in objects[:]])
        print 'last object:', objects[len(objects) - 1]
        print 'position[4]:', position[4]

    new_index = start + 9 * find_triangles(value0, value1, value2, value3, value4, value5, value6, value7, x, x1, y, y1, z, z1, &tridata[start])
    if store_areas:
        areas[i, j, k] = _tri_area(tridata, start, new_index)
    return new_index

# a matrix whose values are unambiguously outside the object
big_number_matrix = numpy.zeros([4, 4, 4])
for i in range(4):
    for j in range(4):
        for k in range(4):
            big_number_matrix[i, j, k] = 9999 + i * 16 + j * 4 + k

@cython.boundscheck(False)
@cython.wraparound(False)
def volume_inside_cell(int i, int j, int k, list objects, numpy.ndarray[numpy.float_t, ndim=1] xs, numpy.ndarray[numpy.float_t, ndim=1] ys, numpy.ndarray[numpy.float_t, ndim=1] zs):
    cdef double x0 = xs[i], y0 = ys[j], z0 = zs[k]
    cdef double x1 = xs[i + 1], y1 = ys[j + 1], z1 = zs[k + 1]
    cdef double dx = x1 - x0, dy = y1 - y0, dz = z1 - z0
    cdef numpy.ndarray[numpy.float_t, ndim=1] local_xs = numpy.array([x0 - dx, x0, x1, x0 + 2 * dx])
    cdef numpy.ndarray[numpy.float_t, ndim=1] local_ys = numpy.array([y0 - dy, y0, y1, y0 + 2 * dy])
    cdef numpy.ndarray[numpy.float_t, ndim=1] local_zs = numpy.array([z0 - dz, z0, z1, z0 + 2 * dz])
    # big numbers unambiguously outside, no two of which are the same so (probably) no risk of dividing by zero?
    cdef numpy.ndarray[numpy.float_t, ndim=3] data = numpy.array(big_number_matrix)
    cdef double x, y, z
    cdef int i1, j1, k1
    if not objects:
        print 'grr... it thinks there is surface when no nearby objects.'
    for i, x in zip([1, 2], [x0, x1]):
        for j, y in zip([1, 2], [y0, y1]):
            for k, z in zip([1, 2], [z0, z1]):
                data[i, j, k] = min([objdist(x, y, z) for objdist in objects])

    # TODO: I probably don't need the 8 corner voxels
    # 9 doubles/tri * 5 tri/cell * 27 cells = 1215 doubles
    cdef numpy.ndarray[numpy.float_t, ndim=1] tridata = numpy.zeros(1215)
    cdef int start = 0
    for i in range(3):
        i1 = i + 1
        for j in range(3):
            j1 = j + 1
            for k in range(3):
                k1 = k + 1
                x, y, z = xs[i], ys[j], zs[k]
                x1, y1, z1 = xs[i + 1], ys[j + 1], zs[k + 1]
                value0, value1, value2, value3, value4, value5, value6, value7 = (data[i, j, k],
                                                                                  data[i1, j, k],
                                                                                  data[i1, j1, k],
                                                                                  data[i, j1, k],
                                                                                  data[i, j, k1],
                                                                                  data[i1, j, k1],
                                                                                  data[i1, j1, k1],
                                                                                  data[i, j1, k1])
                # TODO: if i = 0, j = k = 1, then this finds the triangles to the left (so can get area that way), etc...
                start += 9 * find_triangles(value0, value1, value2, value3, value4, value5, value6, value7, x, x1, y, y1, z, z1, &tridata[start])
    # at this point tridata contains a triangularization that encloses the entire cell's volume
    tridata.resize(start, refcheck=False)
    return tri_volume(tridata)


cdef append_with_deltas(list cell_list, int i, int j, int k):
    cdef int im1 = i - 1, jm1 = j - 1, km1 = k - 1, ip1 = i + 1, jp1 = j + 1, kp1 = k + 1
    cell_list += [(im1, jm1, km1), (im1, jm1, k), (im1, jm1, kp1), (im1, j, km1),
                  (im1, j, k), (im1, j, kp1), (im1, jp1, km1), (im1, jp1, k),
                  (im1, jp1, kp1), (i, jm1, km1), (i, jm1, k), (i, jm1, kp1),
                  (i, j, km1), (i, j, kp1), (i, jp1, km1), (i, jp1, k), (i, jp1, kp1),
                  (ip1, jm1, km1), (ip1, jm1, k), (ip1, jm1, kp1), (ip1, j, km1),
                  (ip1, j, k), (ip1, j, kp1), (ip1, jp1, km1), (ip1, jp1, k),
                  (ip1, jp1, kp1)]









# TODO: move this someplace else. also useful for voxelize
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef chunkify(objects, xs, ys, zs, int chunk_size, double dx):

    cdef int almost = chunk_size - 1
    cdef int nx = (len(xs) + almost) // chunk_size
    cdef int ny = (len(ys) + almost) // chunk_size
    cdef int nz = (len(zs) + almost) // chunk_size
    cdef list objects_distances = [obj.distance for obj in objects]
    cdef int m, i, j, k

    # this is bigger than sqrt(3) * dx / 2 \approx 0.866, the distance from center of cube to corner
    cdef double max_d = dx * .87 * chunk_size

    # safety factor
    max_d *= 2

    cdef list chunk_objs = [[[[] for k in range(nz)] for j in range(ny)] for i in range(nx)]

    # identify chunk_objs
    cdef int lenx = len(xs)
    cdef int leny = len(ys)
    cdef int lenz = len(zs), robj = 0
    cdef double xlo, xhi, ylo, yhi, zlo, zhi
    cdef double bufferdx = 3 * dx

    # TODO: the is_skew_cone business is here because distances are not the real distances in that case;
    #       remove it when I fix this (and will get better performance)
    for m, obj in enumerate(objects):
        is_skew_cone = isinstance(obj, graphicsPrimitives.SkewCone)
        objdist = obj.distance
        for i in range(nx):
            xlo, xhi = xs[i * chunk_size], xs[min((i + 1) * chunk_size - 1, lenx - 1)]
            chunk_objsi = chunk_objs[i]
            # CTNG:testbb
            if not obj.overlaps_x(xlo - bufferdx, xhi + bufferdx): continue
            for j in range(ny):
                ylo, yhi = ys[j * chunk_size], ys[min(leny - 1, (j + 1) * chunk_size - 1)]
                if not obj.overlaps_y(ylo - bufferdx, yhi + bufferdx): continue
                for k in range(nz):
                    zlo, zhi = zs[k * chunk_size], zs[min(lenz - 1, (k + 1) * chunk_size - 1)]
                    if not obj.overlaps_z(zlo - bufferdx, zhi + bufferdx): continue
                    # CTNG:testball
                    if is_skew_cone or objdist((xlo + xhi) * .5, (ylo + yhi) * .5, (zlo + zhi) * .5) < max_d:
                        chunk_objsi[j][k].append(objdist)
    return chunk_objs, nx, ny, nz










# This assumes dx = dy = dz
# CTNG:constructivecubes
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef triangulate_surface(list objects, xs, ys, zs, internal_membranes):
    # use chunks no smaller than 10 voxels across, but aim for max_chunks chunks
    cdef int chunk_size = max(10, int((len(xs) * len(ys) * len(zs) / max_chunks) ** (1 / 3.)))
    cdef double grid_dx = xs[1] - xs[0], grid_dy = ys[1] - ys[0], grid_dz = zs[1] - zs[0]
    cdef double dx = grid_dx

    cdef list chunk_objs
    cdef int nx, ny, nz
    chunk_objs, nx, ny, nz = chunkify(objects, xs, ys, zs, chunk_size, dx)
    return _triangulate_surface_given_chunks(objects, xs, ys, zs, internal_membranes, chunk_size, chunk_objs, nx, ny, nz, False, None)

cpdef _triangulate_surface_given_chunks(list objects, xs, ys, zs, internal_membranes, int chunk_size, list chunk_objs, int nx, int ny, int nz, bint store_areas, areas):
    cdef int i, j, k, di, dj, dk
    cdef double area
    cdef double grid_dx = xs[1] - xs[0], grid_dy = ys[1] - ys[0], grid_dz = zs[1] - zs[0]
    # TODO: we assume a cubic mesh
    # assert(grid_dx == grid_dy == grid_dz)
    cdef double r_inner = grid_dx / 2., r_outer = r_inner * sqrt(3)
    cdef double dx = grid_dx

    cdef list cell_list

    cdef dict to_process = {}

    cdef int cell_count = 0
    cdef int dup_count = 0
    cdef int last_starti

    cdef list cell_list2 = []

    cdef numpy.ndarray[numpy.float_t, ndim=1] triangles
    cdef int surf_count = 0

    cdef dict cur_processed
    cdef int brute_force_count = 0
    cdef list clip_objs
    cdef tuple cell_id
    cdef int new_index

    cdef int numcompartments
    # locate all the potential boundary locations
    triangles = numpy.zeros(1000)
    cdef int triangles_i = 0
    cdef bint reject_if_outside
    for m, obj in enumerate(objects):
        # TODO: remove all the stuff about reject_if_outside when have true SkewCone distances
        reject_if_outside = not(isinstance(obj, graphicsPrimitives.SkewCone))
        objdist = obj.distance
        clip_objs = sum([clipper.primitives for clipper in obj.get_clip()], [])
        cell_list = []
        cur_processed = {}
        for i, j, k in obj.starting_points(xs, ys, zs):
            for di, dj, dk in itertools.product([0, -1, 1, -2, 2], [0, -1, 1, -2, 2], [0, -1, 1, -2, 2]):
                if contains_surface(i + di, j + dj, k + dk, objdist, xs, ys, zs, grid_dx, r_inner, r_outer, reject_if_outside):
                    cell_list.append((i + di, j + dj, k + dk))
                    break
            if cell_list: break
        else:
            # CTNG:systemsearch
            brute_force_count += 1
            # TODO: we are highly unlikely to ever reach this code (happened 0 times with the neuron I tested)
            #       can we prove we never get here?

            numcompartments = (4 + len(range(bisect.bisect_left(xs, obj.xlo), bisect.bisect_right(xs, obj.xhi)))) * (4 + len(range(bisect.bisect_left(ys, obj.ylo), bisect.bisect_right(ys, obj.yhi)))) * (4 + len(range(bisect.bisect_left(zs, obj.zlo), bisect.bisect_right(zs, obj.zhi))))
            for i, j, k in itertools.product(xrange(bisect.bisect_left(xs, obj.xlo) - 2, bisect.bisect_right(xs, obj.xhi) + 2), xrange(bisect.bisect_left(ys, obj.ylo) - 2, bisect.bisect_right(ys, obj.yhi) + 2), xrange(bisect.bisect_left(zs, obj.zlo) - 2, bisect.bisect_right(zs, obj.zhi) + 2)):
                if contains_surface(i, j, k, objdist, xs, ys, zs, grid_dx, r_inner, r_outer, reject_if_outside):
                    cell_list.append((i, j, k))
                    break
            else:
                #print 'objects with no contribution to surface'
                pass

        if not internal_membranes:
            # CTNG:floodfill
            while cell_list:
                cell_id = cell_list.pop()
                if cell_id not in cur_processed:
                    cur_processed[cell_id] = 0
                    i, j, k = cell_id
                    if contains_surface(i, j, k, objdist, xs, ys, zs, grid_dx, r_inner, r_outer, reject_if_outside):
                        to_process[cell_id] = 0
                        append_with_deltas(cell_list, i, j, k)

        else:
            # TODO: is this support ever really useful?
            while cell_list:
                cell_id = cell_list.pop()
                if cell_id not in cur_processed:
                    cur_processed[cell_id] = 0
                    i, j, k = cell_id
                    if contains_surface(i, j, k, objdist, xs, ys, zs, grid_dx, r_inner, r_outer, reject_if_outside):
                        if triangles_i > triangles.size - 50:
                            triangles.resize(triangles.size * 2, refcheck=False)
                        triangles_i = process_cell(i, j, k, [objdist], xs, ys, zs, triangles, triangles_i, store_areas=store_areas, areas=areas)
                        append_with_deltas(cell_list, i, j, k)

    if internal_membranes:
        triangles.resize(triangles_i, refcheck=False)
        return triangles

    cur_processed = None
    cdef list chunk_pts = [[[[] for k in range(nz)] for j in range(ny)] for i in range(nx)]




    # identify chunk_pts
    for i, j, k in to_process.keys():
        chunk_pts[i // chunk_size][j // chunk_size][k // chunk_size].append((i, j, k))

    cdef int num_keys = len(to_process.keys())
    cdef int missing_objs = 0

    triangles = numpy.zeros(45 * num_keys)
    cdef int starti = 0
    cdef int a, b, c
    cdef int extra_count
    cdef list obj_search
    # handle a chunk at a time
    for a in range(nx):
        chunk_objsa = chunk_objs[a]
        chunk_ptsa = chunk_pts[a]
        for b in range(ny):
            for c in range(nz):
                objs = chunk_objsa[b][c]
                cells = chunk_ptsa[b][c]
                if cells and not objs:
                    # we should never get here; this is just a sanity check
                    missing_objs += 1
                    continue
                for i, j, k in cells:
                    last_starti = starti
                    #starti = process_cell(i, j, k, objects_distances, xs, ys, zs, triangles, last_starti) #was objs
                    #tri_data = list(triangles[last_starti : starti])
                    start_i = process_cell(i, j, k, objs, xs, ys, zs, triangles, last_starti, store_areas=store_areas, areas=areas)
                    '''
                    # this an the two-above commented-out lines are for debugging to detect discrepancies between the
                    # chunked partitioning and using all the nodes
                    if starti - last_starti != len(tri_data) or any(tri_data != triangles[last_starti : starti]):
                        print 'discrepancy in grid (%d, %d, %d) -- chunk_size = %d' % (i, j, k, chunk_size)
                        # identify which object(s)... grr...
                        """
                        for m in xrange(len(objects)):
                            tri_data2 = list(triangles[last_starti : starti])
                            starti = process_cell(i, j, k, objs + objects_distances[: m], xs, ys, zs, triangles, last_starti) #was objs
                            if starti - last_starti != len(tri_data2) or any(tri_data2 != triangles[last_starti : starti]):
                                print '   missed object %d: %r' % (m, objects[m - 1])
                            if starti - last_starti == len(tri_data) and all(tri_data == triangles[last_starti : starti]):
                                break
                        else:
                            print '    *** should never get here ***'"""
                    '''

    # some regions of the surface may yet be missing
    # it's possible for voxels to not contain detectable surface of any component, but
    # contain detectable surface for their union
    # to find, we look for holes in the surface and flood from them, checking against every object
    # TODO: could be smarter about this: at least check (enlarged) bounding boxes first


    last_starti = 0
    cdef dict process2
    pt_neighbor_map = {}
    cdef dict count
    cdef int old_start_i
    cdef list still_to_process, local_objs
    # append to the pt_neighbor_map
    process2 = {}
    count = {}

    for i in xrange(last_starti, starti, 9):
        pts = {}
        for j in xrange(3):
            pts[tuple(triangles[i + 3 * j : i + 3 * j + 3])] = 0
        pts = list(pts.keys())
        if len(pts) == 3:
            # only consider triangles of nonzero area
            for j in xrange(3):
                for k in xrange(3):
                    if j != k:
                        _register_on_neighbor_map(pt_neighbor_map, pts[j], pts[k])

    last_starti = starti
    xlo, ylo, zlo = xs[0], ys[0], zs[0]

    # if no holes, each point should have each neighbor listed more than once
    for pt, neighbor_list in zip(pt_neighbor_map.keys(), pt_neighbor_map.values()):
        count = {}
        for neighbor in neighbor_list:
            if neighbor not in count:
                count[neighbor] = 1
            else:
                count[neighbor] += 1
        for neighbor, ncount in zip(count.keys(), count.values()):
            if ncount <= 1:
                # Note: this assumes (as we stated above) that dx = dy = dz
                for point in (pt, neighbor):
                    i, j, k = (point[0] - xlo) // dx, (point[1] - ylo) // dx, (point[2] - zlo) // dx
                    for di in range(-1, 2):
                        for dj in range(-1, 2):
                            for dk in range(-1, 2):
                                cell_id = (i + di, j + dj, k + dk)
                                if cell_id not in process2 and cell_id not in to_process:
                                    process2[cell_id] = 0
                break

    still_to_process = list(process2.keys())
    #print 'len(still_to_process) = %d' % len(still_to_process)
    # flood on those still_to_process
    while still_to_process:
        cell_id = still_to_process.pop()
        # make sure we haven't already handled this voxel
        if cell_id not in to_process:
            i, j, k = cell_id
            old_start_i = starti
            local_objs = chunk_objs[i // chunk_size][j // chunk_size][k // chunk_size]
            if local_objs:
                for m, objdist in enumerate(local_objs):
                    if contains_surface(i, j, k, objdist, xs, ys, zs, dx, r_inner, r_outer, False):
                        print 'item %d in grid(%d, %d, %d) contains previously undetected surface' % (m, i, j, k)
                        for n, obj in enumerate(objects):
                            if obj.distance == objdist:
                                print '    (i.e. global item %d: %r)' % (n, obj)
                                break
                #starti = process_cell(i, j, k, local_objs, xs, ys, zs, triangles, starti) # was objects_distances
                starti = process_cell(i, j, k, local_objs, xs, ys, zs, triangles, starti, store_areas=store_areas, areas=areas)

            # mark it off so we know we don't visit that grid point again
            to_process[cell_id] = 0
            if old_start_i != starti:
                append_with_deltas(still_to_process, i, j, k)

    triangles.resize(starti, refcheck=False)

    return triangles


cpdef double tri_area(numpy.ndarray[numpy.float_t, ndim=1] triangles):
    return _tri_area(triangles, 0, len(triangles))


# CTNG:surfacearea
@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double _tri_area(numpy.ndarray[numpy.float_t, ndim=1] triangles, int lo, int hi):
    cdef double doublearea = 0., local_area
    cdef int i
    for i in range(lo, hi, 9):
        local_area = llgramarea(&triangles[i], &triangles[3 + i], &triangles[6 + i])
        doublearea += local_area
        if numpy.isnan(local_area):
            print 'tri_area exception: ', ', '.join([str(v) for v in triangles[i : i + 9]])
    return doublearea * 0.5

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef double tri_volume(numpy.ndarray[numpy.float_t, ndim=1] triangles):
    cdef double sixtimesvolume = 0., local_vol
    cdef int i
    for i in range(0, len(triangles), 9):
        local_vol = llpipedfromoriginvolume(&triangles[i], &triangles[3 + i], &triangles[6 + i])
        sixtimesvolume += local_vol
        if numpy.isnan(local_vol):
            print 'tri_volume exception:',
            for j in range(i, i + 9):
                print triangles[j],
                print
    return abs(sixtimesvolume / 6.)



from neuron import h, crxd as rxd
from neuron.crxd import v
from neuron.crxd.rxdmath import vtrap, exp, log
from math import pi

h.load_file("stdrun.hoc")
h.CVode().active(True)
# parameters
h.celsius = 6.3
e = 1.60217662e-19
scale = 1e-14 / e
gnabar = 0.12 * scale  # molecules/um2 ms mV
gkbar = 0.036 * scale
gl = 0.0003 * scale
el = -54.3
q10 = 3.0 ** ((h.celsius - 6.3) / 10.0)

# sodium activation 'm'
alpha = 0.1 * vtrap(-(v + 40.0), 10)
beta = 4.0 * exp(-(v + 65) / 18.0)
mtau = 1.0 / (q10 * (alpha + beta))
minf = alpha / (alpha + beta)


# sodium inactivation 'h'
alpha = 0.07 * exp(-(v + 65.0) / 20.0)
beta = 1.0 / (exp(-(v + 35.0) / 10.0) + 1.0)
htau = 1.0 / (q10 * (alpha + beta))
hinf = alpha / (alpha + beta)

# potassium activation 'n'
alpha = 0.01 * vtrap(-(v + 55.0), 10.0)
beta = 0.125 * exp(-(v + 65.0) / 80.0)
ntau = 1.0 / (q10 * (alpha + beta))
ninf = alpha / (alpha + beta)

somaA = h.Section("somaA")
somaA.pt3dclear()
somaA.pt3dadd(-90, 0, 0, 30)
somaA.pt3dadd(-60, 0, 0, 30)
somaA.nseg = 1

somaB = h.Section("somaB")
somaB.pt3dclear()
somaB.pt3dadd(60, 0, 0, 30)
somaB.pt3dadd(90, 0, 0, 30)
somaB.nseg = 1

# mod version
somaB.insert("hh")

# rxd version

# Where?
# intracellular
cyt = rxd.Region(h.allsec(), name="cyt", nrn_region="i")

# membrane
mem = rxd.Region(h.allsec(), name="cell_mem", geometry=rxd.membrane())

# extracellular
ecs = rxd.Extracellular(-100, -100, -100, 100, 100, 100, dx=33)

# Who?
def init(ics, ecs):
    return lambda nd: ecs if isinstance(nd, rxd.node.NodeExtracellular) else ics


# ions
k = rxd.Species(
    [cyt, mem], name="k", d=1, charge=1, initial=54.4, represents="CHEBI:29103"
)

kecs = rxd.Parameter([ecs], name="k", value=2.5, charge=1, represents="CHEBI:29103")

na = rxd.Species(
    [cyt, mem], name="na", d=1, charge=1, initial=10.0, represents="CHEBI:29101"
)

naecs = rxd.Parameter([ecs], name="na", value=140, charge=1, represents="CHEBI:29101")

x = rxd.Species([cyt, mem, ecs], name="x", charge=1, initial=1e9)

ki, ko, nai, nao, xi, xo = k[cyt], kecs[ecs], na[cyt], naecs[ecs], x[cyt], x[ecs]

# gates
ngate = rxd.State([cyt, mem], name="ngate", initial=0.24458654944007166)
mgate = rxd.State([cyt, mem], name="mgate", initial=0.028905534475191907)
hgate = rxd.State([cyt, mem], name="hgate", initial=0.7540796658225246)

# somaA parameter
pA = rxd.Parameter(
    [cyt, mem], name="paramA", initial=lambda nd: 1 if nd.segment in somaA else 0
)

# What
# gates
m_gate = rxd.Rate(mgate, (minf - mgate) / mtau)
h_gate = rxd.Rate(hgate, (hinf - hgate) / htau)
n_gate = rxd.Rate(ngate, (ninf - ngate) / ntau)

# Nernst potentials
ena = 1e3 * h.R * (h.celsius + 273.15) * log(nao / nai) / h.FARADAY
ek = 1e3 * h.R * (h.celsius + 273.15) * log(ko / ki) / h.FARADAY

gna = pA * gnabar * mgate**3 * hgate
gk = pA * gkbar * ngate**4

na_current = rxd.MultiCompartmentReaction(
    nai, nao, gna * (v - ena), mass_action=False, membrane=mem, membrane_flux=True
)
k_current = rxd.MultiCompartmentReaction(
    ki, ko, gk * (v - ek), mass_action=False, membrane=mem, membrane_flux=True
)
leak_current = rxd.MultiCompartmentReaction(
    xi, xo, pA * gl * (v - el), mass_action=False, membrane=mem, membrane_flux=True
)

# stimulate
stimA = h.IClamp(somaA(0.5))
stimA.delay = 50
stimA.amp = 1
stimA.dur = 50

stimB = h.IClamp(somaB(0.5))
stimB.delay = 50
stimB.amp = 1
stimB.dur = 50

# record
tvec = h.Vector().record(h._ref_t)

vvecA = h.Vector().record(somaA(0.5)._ref_v)
mvecA = h.Vector().record(mgate.nodes(somaA(0.5))[0]._ref_value)
nvecA = h.Vector().record(ngate.nodes(somaA(0.5))[0]._ref_value)
hvecA = h.Vector().record(hgate.nodes(somaA(0.5))[0]._ref_value)
kvecA = h.Vector().record(somaA(0.5)._ref_ik)
navecA = h.Vector().record(somaA(0.5)._ref_ina)

vvecB = h.Vector().record(somaB(0.5)._ref_v)
kvecB = h.Vector().record(somaB(0.5)._ref_ik)
navecB = h.Vector().record(somaB(0.5)._ref_ina)
mvecB = h.Vector().record(somaB(0.5).hh._ref_m)
nvecB = h.Vector().record(somaB(0.5).hh._ref_n)
hvecB = h.Vector().record(somaB(0.5).hh._ref_h)
tvec = h.Vector().record(h._ref_t)

# run
h.finitialize(-70)
# for i in range(1000):
#    h.fadvance()
#    print(h.t,somaA(0.5).v, somaA(0.5).ki, somaA(0.5).nai, somaA(0.5).xi)
h.continuerun(100)


if __name__ == "__main__":
    from matplotlib import pyplot

    # plot the results
    pyplot.ion()
    fig = pyplot.figure()
    pyplot.plot(tvec, vvecA, label="rxd")
    pyplot.plot(tvec, vvecB, label="mod")
    pyplot.legend()
    fig.set_dpi(200)

    fig = pyplot.figure()
    pyplot.plot(tvec, hvecA, "-b", label="h")
    pyplot.plot(tvec, mvecA, "-r", label="m")
    pyplot.plot(tvec, nvecA, "-g", label="n")
    pyplot.plot(tvec, hvecB, ":b")
    pyplot.plot(tvec, mvecB, ":r")
    pyplot.plot(tvec, nvecB, ":g")
    pyplot.legend()
    fig.set_dpi(200)

    fig = pyplot.figure()
    pyplot.plot(tvec, kvecA.as_numpy(), "-b", label="k")
    pyplot.plot(tvec, navecA.as_numpy(), "-r", label="na")
    pyplot.plot(tvec, kvecB, ":b")
    pyplot.plot(tvec, navecB, ":r")
    pyplot.legend()
    fig.set_dpi(200)



"""
Collection of all unittests for the neuron module

$Id$
"""


# import your specific test here
# and add it to the "suite" function below
from neuron.tests import test_vector, test_neuron

import unittest


def suite():

    suite = unittest.TestSuite()
    suite.addTest(test_vector.suite())
    suite.addTest(test_neuron.suite())
    # add additional test cases here
    return suite


if __name__ == "__main__":

    # unittest.main()
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(suite())



from neuron import h, crxd as rxd
import numpy
import __main__

name = __main__.__file__
if name[-3:] == ".py":
    name = name[:-3]
h.load_file("stdrun.hoc")

dend1 = h.Section()
dend1.diam = 2
dend1.nseg = 101
dend1.L = 50

dend2 = h.Section()
dend2.diam = 2
dend2.nseg = 101
dend2.L = 50
dend2.connect(dend1)

diff_constant = 1
h.cvode_active(True)
r = rxd.Region(h.allsec(), dx=0.25)

rxd.set_solve_type([dend1], dimension=3)

ca = rxd.Species(
    r,
    d=diff_constant,
    atolscale=0.1,
    initial=lambda node: 1
    if (0.8 < node.x and node.segment in dend1)
    or (node.x < 0.2 and node.segment in dend2)
    else 0,
)
bistable_reaction = rxd.Rate(ca, -ca * (1 - ca) * (0.01 - ca))
h.finitialize()

for i in range(2):
    h.fadvance()
rxd.nthread(2)
for i in range(2, 4):
    h.fadvance()
rxd.nthread(3)
for i in range(4, 6):
    h.fadvance()
rxd.nthread(4)
for i in range(6, 8):
    h.fadvance()
rxd.nthread(1)
for i in range(8, 11):
    h.fadvance()



from .xml2nrn import *


def cable(self, node):
    self.lastcabid_ = -1
    if self.in_cablegroup_:
        self.cablegroups_[-1].cable_indices_.append(
            self.cableid2index_[int(node.get("id"))]
        )
    else:
        cab = self.cables_[self.cableid2index_[int(node.get("id"))]]
        self.lastcabid_ = int(node.get("id"))
        val = node.get("fractAlongParent")
        if val is not None:
            cab.px_ = float(val)
        else:
            cab.px_ = 1.0
        cab.name_ = str(node.get("name"))


def cablegroup(self, node):
    self.in_cablegroup_ = True
    name = str(node.get("name"))
    self.cablegroups_.append(CableGroup(name))
    self.groupname2index_[name] = len(self.cablegroups_) - 1


def cables(self, node):
    pass


def distal(self, node):
    pt = Point(self.id, self.pid, self.cid, self.locator.getLineNumber())
    self.ptid2pt_[self.id] = pt
    pt.set(
        float(node.get("x")),
        float(node.get("y")),
        float(node.get("z")),
        float(node.get("diameter")),
    )
    if self.cable_.pcnt_ == 1:
        proxpt = self.points_[len(self.points_) - 1]
        if (
            proxpt.x_ == pt.x_
            and proxpt.y_ == pt.y_
            and proxpt.z_ == pt.z_
            and proxpt.d_ == pt.d_
        ):
            if debug:
                print("Prox and distal same, assuming spherical segment")
            pt.y_ = pt.y_ + (pt.d_ / 2.0)
            self.points_[len(self.points_) - 1].y_ = self.points_[
                len(self.points_) - 1
            ].y_ - (self.points_[len(self.points_) - 1].d_ / 2.0)
            if debug:
                print("New distal: " + str(pt))
                print("New proximal: " + str(self.points_[len(self.points_) - 1]))
    self.points_.append(pt)
    self.cable_.pcnt_ += 1
    if debug:
        print("Distal: " + str(pt))
        print("Cable ", self.cable_.id_, " has ", self.cable_.pcnt_, " points")


def proximal(self, node):
    self.nprox += 1
    pt = Point(-1, self.pid, self.cid, self.locator.getLineNumber())
    pt.set(
        float(node.get("x")),
        float(node.get("y")),
        float(node.get("z")),
        float(node.get("diameter")),
    )
    self.points_.append(pt)
    self.cable_.pcnt_ += 1
    if debug:
        print("Proximal: " + str(pt))
        print("Cable ", self.cable_.id_, " has ", self.cable_.pcnt_, " points")


def segment(self, node):
    self.id = int(node.get("id"))
    self.cid = int(node.get("cable"))
    parent_cable_id = -1
    p = node.get("parent")
    if p is not None:
        self.pid = int(p)
        parent_cable_id = self.ptid2pt_[self.pid].cid_
    else:
        self.pid = -1

    if debug:
        print(
            "\nsegment id=",
            self.id,
            "  cable=",
            self.cid,
            " parent id=",
            self.pid,
            " parent_cable_id=",
            parent_cable_id,
        )

    if self.cable_ is None:
        self.cable_ = Cable(self.cid, self.pid, len(self.points_))
        self.cableid2index_[self.cid] = len(self.cables_)
        self.cables_.append(self.cable_)
        self.cable_.parent_cable_id_ = parent_cable_id
    if self.cable_.id_ != self.cid:
        self.cable_ = Cable(self.cid, self.pid, len(self.points_))
        self.cableid2index_[self.cid] = len(self.cables_)
        self.cables_.append(self.cable_)
        self.cable_.parent_cable_id_ = parent_cable_id


def segments(self, node):
    self.in_cablegroup_ = False
    self.points_ = []
    self.cables_ = []
    self.cable_ = None
    self.id = -1
    self.cid = -1
    self.pid = -1
    self.nprox = 0
    self.cableid2index_ = {}
    self.ptid2pt_ = {}
    self.cablegroups_ = []
    self.groupname2index_ = {}


def segments_end(self, node):
    if debug:
        print("\nEnd of segments element")
        ic = 0
        ip = 0
        for cab in self.cables_:
            ic += 1
            for i in range(cab.first_, cab.first_ + cab.pcnt_):
                pt = self.points_[i]
                print(ip, pt.id_, pt.pid_, pt.x_, pt.y_, pt.z_, pt.d_)
                ip += 1
        print("ncable=", ic, "  npoint=", ip, "   nprox=", self.nprox, "\n")

    return



#cython: language_level=2
# this contains cone, spherecone, and cylinder code translated and modified from Barbier and
# Galin 2004's example code
# see /u/ramcd/spatial/experiments/one_time_tests/2012-06-28/cone.cpp
# on 7 june 2012, the original code was available online at
#   http://jgt.akpeters.com/papers/BarbierGalin04/Cone-Sphere.zip

import bisect
cimport cython

"""
The graphicsPrimitive module
"""

cdef extern from "math.h":
    double sqrt(double)
    double fabs(double)

# I'm manually doing all maxes and mins here because I thought it would help
# eliminate round-off issues at the surface. it doesn't

#cdef double max(double a, double b):
#    return a if a > b else b

#cdef double min(double a, double b):
#    return a if a < b else b

cdef class Complement:
    def __init__(self, obj):
        self.obj = obj
    def __repr__(self):
        return 'Complement(%r)' % self.obj
    def distance(self, px, py, pz):
        return -self.obj.distance(px, py, pz)
    def starting_points(self, xs, ys, zs):
        return self.obj.starting_points(xs, ys, zs)
    property primitives:
        def __get__(self): return [self.obj]


cdef class Union:
    cdef list objects
    def __init__(self, list objects):
        self.objects = objects
    def __repr__(self):
        return 'Union(%r)' % self.objects
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef double distance(self, px, py, pz):
        # CTNG:uniondist
        dists = [obj.distance(px, py, pz) for obj in self.objects]
        cdef double d, d2
        d = dists[0]
        for d2 in dists:
            if d2 < d: d = d2
        return d
    cpdef list starting_points(self, xs, ys, zs):
        return sum([c.starting_points(xs, ys, zs) for c in self.objects], [])
    property primitives:
        def __get__(self): return sum([obj.primitives for obj in self.objects], [])


cdef class Intersection:
    cdef list objects
    def __init__(self, list objects):
        self.objects = objects
    def __repr__(self):
        return 'Intersection(%r)' % self.objects
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef double distance(self, px, py, pz):
        # CTNG: intdist
        dists = [obj.distance(px, py, pz) for obj in self.objects]
        cdef double d, d2
        d = dists[0]
        for d2 in dists:
            if d2 > d: d = d2
        return d
    cpdef list starting_points(self, xs, ys, zs):
        return sum([c.starting_points(xs, ys, zs) for c in self.objects], [])
    property primitives:
        def __get__(self): return sum([obj.primitives for obj in self.objects], [])



cdef class Plane:
    cdef double d, mul, nx, ny, nz, px, py, pz
    def __init__(self, double px, double py, double pz, double nx, double ny, double nz):
        """(px, py, pz) -- a point; (nx, ny, nz) -- the normal vector"""
        self.d = - (nx * px + ny * py + nz * pz)
        self.mul = 1. / sqrt(nx ** 2 + ny ** 2 + nz ** 2)
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.px = px
        self.py = py
        self.pz = pz
    property primitives:
        def __get__(self): return []
    def __repr__(self):
        return 'Plane(%g, %g, %g, %g, %g, %g)' % (self.px, self.py, self.pz, self.nx, self.ny, self.nz)
    cpdef double distance(self, double x, double y, double z):
        return (self.nx * x + self.ny * y + self.nz * z + self.d) * self.mul
    cpdef list starting_points(self, xs, ys, zs):
        return [(bisect.bisect_left(xs, self.px), bisect.bisect_left(ys, self.py), bisect.bisect_left(zs, self.pz))]


cdef class Sphere:
    cdef double _x, _y, _z, _r
    cdef _xlo, _xhi, _ylo, _yhi, _zlo, _zhi
    cdef list clips
    property primitives:
        def __get__(self): return [self]

    def __init__(self, double x, double y, double z, double r):
        self._x, self._y, self._z, self._r = x, y, z, r
        self._xlo, self._xhi = x - r, x + r
        self._ylo, self._yhi = y - r, y + r
        self._zlo, self._zhi = z - r, z + r
        self.clips = []
    def __repr__(self):
        if self.clips:
            return 'Sphere(%g, %g, %g, %g; clips=%r)' % (self._x, self._y, self._z, self._r, self.clips)
        else:
            return 'Sphere(%g, %g, %g, %g)' % (self._x, self._y, self._z, self._r)
    property x:
        def __get__(self): return self._x
    property y:
        def __get__(self): return self._y
    property z:
        def __get__(self): return self._z
    property r:
        def __get__(self): return self._r
    property xlo:
        def __get__(self): return self._xlo
    property xhi:
        def __get__(self): return self._xhi
    property ylo:
        def __get__(self): return self._ylo
    property yhi:
        def __get__(self): return self._yhi
    property zlo:
        def __get__(self): return self._zlo
    property zhi:
        def __get__(self): return self._zhi
    cpdef double distance(self, double x, double y, double z):
        d = sqrt((x - self._x) ** 2 + (y - self._y) ** 2 + (z - self._z) ** 2) - self._r
        for clip in self.clips:
            d = max(d, clip.distance(x, y, z))
        return d
    def starting_points(self, xs, ys, zs):
        #for theta in numpy.arange(0, 2 * numpy.pi, 10):
        # TODO: this only works right if the entire object is inside the domain
        return sum([c.starting_points(xs, ys, zs) for c in self.clips], [(bisect.bisect_left(xs, self._x - self._r), bisect.bisect_left(ys, self._y), bisect.bisect_left(zs, self._z))])
    cpdef bint overlaps_x(self, double lo, double hi):
        return lo <= self._xhi and hi >= self._xlo
    cpdef bint overlaps_y(self, double lo, double hi):
        return lo <= self._yhi and hi >= self._ylo
    cpdef bint overlaps_z(self, double lo, double hi):
        return lo <= self._zhi and hi >= self._zlo
    def set_clip(self, clips):
        self.clips = clips
    def get_clip(self):
        return self.clips



cdef class Cylinder:
    cdef double cx, cy, cz, r, rr, axisx, axisy, axisz, x0, y0, z0, x1, y1, z1, h
    cdef double length, _xlo, _xhi, _ylo, _yhi, _zlo, _zhi
    cdef list neighbors, clips, neighbor_regions
    def __repr__(self):
        if self.clips:
            return 'Cylinder(%g, %g, %g, %g, %g, %g, %g; clips=%r)' % (self.x0, self.y0, self.z0, self.x1, self.y1, self.z1, self.r, self.clips)
        else:
            return 'Cylinder(%g, %g, %g, %g, %g, %g, %g)' % (self.x0, self.y0, self.z0, self.x1, self.y1, self.z1, self.r)
    property xlo:
        def __get__(self): return self._xlo
    property xhi:
        def __get__(self): return self._xhi
    property ylo:
        def __get__(self): return self._ylo
    property yhi:
        def __get__(self): return self._yhi
    property zlo:
        def __get__(self): return self._zlo
    property zhi:
        def __get__(self): return self._zhi

    def __init__(self, double x0, double y0, double z0, double x1, double y1, double z1, double r):
        self.cx, self.cy, self.cz = (x0 + x1) * 0.5, (y0 + y1) * 0.5, (z0 + z1) * 0.5
        self.x0, self.y0, self.z0, self.x1, self.y1, self.z1 = x0, y0, z0, x1, y1, z1
        self.rr = r * r
        self.r = r
        self.axisx, self.axisy, self.axisz = (x1 - x0, y1 - y0, z1 - z0)
        self.length = sqrt(self.axisx ** 2 + self.axisy ** 2 + self.axisz ** 2)
        # normalize the axis
        self.axisx /= self.length
        self.axisy /= self.length
        self.axisz /= self.length


        self.h = self.length * 0.5

        self._xlo = min(x0 - r, x1 - r)
        self._xhi = max(x0 + r, x1 + r)
        self._ylo = min(y0 - r, y1 - r)
        self._yhi = max(y0 + r, y1 + r)
        self._zlo = min(z0 - r, z1 - r)
        self._zhi = max(z0 + r, z1 + r)
        self.neighbors = []
        self.clips = []
        self.neighbor_regions = []

    property axislength:
        def __get__(self):
            return self.length

    property _x0:
        def __get__(self):
            return self.x0

    property _y0:
        def __get__(self):
            return self.y0
    property _z0:
        def __get__(self):
            return self.z0
    property _r0:
        def __get__(self):
            return self.r

    property _x1:
        def __get__(self):
            return self.x1

    property _y1:
        def __get__(self):
            return self.y1
    property _z1:
        def __get__(self):
            return self.z1
    property _r1:
        def __get__(self):
            return self.r
    property primitives:
        def __get__(self): return [self]

    def get_clip(self):
        return self.clips


    def set_clip(self, clips):
        self.clips = clips

    def axis(self):
        return (self.axisx, self.axisy, self.axisz)

    def set_neighbors(self, neighbors, neighbor_regions):
        self.neighbors = neighbors
        self.neighbor_regions = neighbor_regions

    cpdef within_core(self, double px, double py, double pz):
        cdef double nx, ny, nz, y, yy, xx
        nx, ny, nz = px - self.cx, py - self.cy, pz - self.cz
        y = abs(self.axisx * nx + self.axisy * ny + self.axisz * nz)
        return y < self.h


    def starting_points(self, xs, ys, zs):
        # TODO: this only works right if the entire object is inside the domain
        return sum([c.starting_points(xs, ys, zs) for c in self.clips],
                    [(bisect.bisect_left(xs, self.x0), bisect.bisect_left(ys, self.y0), bisect.bisect_left(zs, self.z0)),
                     (bisect.bisect_left(xs, self.x1), bisect.bisect_left(ys, self.y1), bisect.bisect_left(zs, self.z1))])


    cpdef double distance(self, double px, double py, double pz):
        # CTNG:cyldist
        cdef double nx, ny, nz, y, yy, xx, d
        nx, ny, nz = px - self.cx, py - self.cy, pz - self.cz
        y = abs(self.axisx * nx + self.axisy * ny + self.axisz * nz)
        yy = y * y
        xx = nx * nx + ny * ny + nz * nz - yy
        if y < self.h:
            # this was wrong in my version from 2012-06-28, since did not
            # handle end-caps being closest
            d = max(y - self.h, sqrt(xx) - self.r)
        else:
            y -= self.h
            if xx < self.rr:
                d = y
            else:
                yy = y * y
                x = sqrt(xx) - self.r
                d = sqrt(yy + x * x)

        for clip in self.clips:
            d = max(d, clip.distance(px, py, pz))
        return d





    cpdef bint overlaps_x(self, double lo, double hi):
        return lo <= self._xhi and hi >= self._xlo
    cpdef bint overlaps_y(self, double lo, double hi):
        return lo <= self._yhi and hi >= self._ylo
    cpdef bint overlaps_z(self, double lo, double hi):
        return lo <= self._zhi and hi >= self._zlo



cdef class SphereCone:
    cdef double x0, y0, z0, r0, x1, y1, z1, r1, rra, rrb, axisx, axisy, axisz, conelength, side1, side2
    cdef double length, _xlo, _xhi, _ylo, _yhi, _zlo, _zhi, ha, hb, hra, hrb
    cdef list clips
    property primitives:
        def __get__(self): return [self]

    def __repr__(self):
        return 'SphereCone(%g, %g, %g, %g, %g, %g, %g, %g)' % (self.x0, self.y0, self.z0, self.r0, self.x1, self.y1, self.z1, self.r1)

    def set_clip(self, clips):
        self.clips = clips
    def get_clip(self):
        return self.clips


    def __init__(self, double x0, double y0, double z0, double r0, double x1, double y1, double z1, double r1):
        if r1 > r0:
            x0, y0, z0, r0, x1, y1, z1, r1 = x1, y1, z1, r1, x0, y0, z0, r0

        self.x0, self.y0, self.z0, self.r0, self.x1, self.y1, self.z1, self.r1 = x0, y0, z0, r0, x1, y1, z1, r1

        self.rra, self.rrb = r0 * r0, r1 * r1
        self.axisx, self.axisy, self.axisz = (x1 - x0, y1 - y0, z1 - z0)
        self.length = sqrt(self.axisx ** 2 + self.axisy ** 2 + self.axisz ** 2)
        # normalize the axis
        self.axisx /= self.length
        self.axisy /= self.length
        self.axisz /= self.length
        """
        rab = r0 - r1
        s = sqrt(self.length ** 2 - rab ** 2)
        self.ha = self.r0 * rab / self.length
        self.hb = self.r1 * rab / self.length
        self.hra = self.r0 * s / self.length
        self.hrb = self.r1 * s / self.length
        self.side1 = - self.ha / self.r0
        self.side2 = self.hra / self.r0
        self.conelength = self.length * self.hra / self.r0
        """
        self.r0 = r0
        self.r1 = r1
        self.rrb = r1 * r1
        self.rra = r0 * r0
        r0b = r0 - r1
        s = sqrt(self.length * self.length - r0b * r0b)
        self.ha = self.r0 * r0b / self.length
        self.hb = self.r1 * r0b / self.length
        self.hra = self.r0 * s / self.length
        self.hrb = self.r1 * s / self.length
        self.side1 = -self.ha / self.r0
        self.side2 = self.hra / self.r0
        self.conelength = self.length * self.hra / self.r0

        self._xlo = min(x0 - r0, x1 - r1)
        self._xhi = max(x0 + r0, x1 + r1)
        self._ylo = min(y0 - r0, y1 - r1)
        self._yhi = max(y0 + r0, y1 + r1)
        self._zlo = min(z0 - r0, z1 - r1)
        self._zhi = max(z0 + r0, z1 + r1)
        self.clips = []

    def starting_points(self, xs, ys, zs):
        # TODO: this only works right if the entire object is inside the domain
        return sum([c.starting_points(xs, ys, zs) for c in self.clips], [(bisect.bisect_left(xs, self.x0 - self.axisx * self.r0), bisect.bisect_left(ys, self.y0 - self.axisy * self.r0), bisect.bisect_left(zs, self.z0 - self.axisz * self.r0))])


    cpdef double distance(self, double px, double py, double pz):
        cdef double nx, ny, nz, y, yy, xx, ry, rx, nn, x, d
        nx, ny, nz = px - self.x0, py - self.y0, pz - self.z0
        y = nx * self.axisx + ny * self.axisy + nz * self.axisz
        nn = nx * nx + ny * ny + nz * nz

        yy = y * y
        xx = nn - yy
        # in principle, xx >= 0, however roundoff errors may cause trouble
        if xx < 0: xx = 0
        x = sqrt(xx)
        ry = x * self.side1 + y * self.side2
        if ry < 0:
            d = sqrt(nn) - self.r0
        elif ry > self.conelength:
            y = y - self.length
            yy = y * y
            nn = xx + yy
            d = sqrt(nn) - self.r1
        else:
            rx = x * self.side2 - y * self.side1
            d = rx - self.r0

        for clip in self.clips:
            d = max(d, clip.distance(px, py, pz))
        return d



    cpdef bint overlaps_x(self, double lo, double hi):
        return lo <= self._xhi and hi >= self._xlo
    cpdef bint overlaps_y(self, double lo, double hi):
        return lo <= self._yhi and hi >= self._ylo
    cpdef bint overlaps_z(self, double lo, double hi):
        return lo <= self._zhi and hi >= self._zlo



cdef class Cone:
    cdef double x0, y0, z0, r0, x1, y1, z1, r1, rra, rrb, axisx, axisy, axisz, conelength, side1, side2
    cdef double length, _xlo, _xhi, _ylo, _yhi, _zlo, _zhi, cx, cy, cz, h
    cdef list neighbors, clips, neighbor_regions
    cdef bint reversed
    property xlo:
        def __get__(self): return self._xlo
    property xhi:
        def __get__(self): return self._xhi
    property ylo:
        def __get__(self): return self._ylo
    property yhi:
        def __get__(self): return self._yhi
    property zlo:
        def __get__(self): return self._zlo
    property zhi:
        def __get__(self): return self._zhi

    def __repr__(self):
        cdef list order
        if self.reversed:
            order = [self.x1, self.y1, self.z1, self.r1, self.x0, self.y0, self.z0, self.r0]
        else:
            order = [self.x0, self.y0, self.z0, self.r0, self.x1, self.y1, self.z1, self.r1]
        if self.clips:
            return 'Cone(%g, %g, %g, %g, %g, %g, %g, %g; clips=%r)' % tuple(order + [self.clips])
        else:
            return 'Cone(%g, %g, %g, %g, %g, %g, %g, %g)' % tuple(order)

    property _x0:
        def __get__(self):
            return self.x1 if self.reversed else self.x0

    property _y0:
        def __get__(self):
            return self.y1 if self.reversed else self.y0
    property _z0:
        def __get__(self):
            return self.z1 if self.reversed else self.z0
    property _r0:
        def __get__(self):
            return self.r1 if self.reversed else self.r0

    property _x1:
        def __get__(self):
            return self.x0 if self.reversed else self.x1

    property _y1:
        def __get__(self):
            return self.y0 if self.reversed else self.y1
    property _z1:
        def __get__(self):
            return self.z0 if self.reversed else self.z1
    property _r1:
        def __get__(self):
            return self.r0 if self.reversed else self.r1

    property axislength:
        def __get__(self):
            return self.length
    property primitives:
        def __get__(self): return [self]


    def __init__(self, double x0, double y0, double z0, double r0, double x1, double y1, double z1, double r1):
        if r1 > r0:
            x0, y0, z0, r0, x1, y1, z1, r1 = x1, y1, z1, r1, x0, y0, z0, r0
            self.reversed = True
        else:
            self.reversed = False

        self.x0, self.y0, self.z0, self.r0, self.x1, self.y1, self.z1, self.r1 = x0, y0, z0, r0, x1, y1, z1, r1

        if r0 < 0:
            raise Exception('At least one Cone radius must be positive')
        if r1 < 0:
            axisx, axisy, axisz = (x1 - x0, y1 - y0, z1 - z0)
            length = sqrt(axisx ** 2 + axisy ** 2 + axisz ** 2)
            axisx /= length; axisy /= length; axisz /= length
            f = r1 / (r1 - r0)
            x1 -= f * axisx; y1 -= f * axisy; z1 -= f * axisz; r1 = 0

        self.rra, self.rrb = r0 * r0, r1 * r1
        self.axisx, self.axisy, self.axisz = (x1 - x0, y1 - y0, z1 - z0)
        self.length = sqrt(self.axisx ** 2 + self.axisy ** 2 + self.axisz ** 2)
        # normalize the axis
        self.axisx /= self.length
        self.axisy /= self.length
        self.axisz /= self.length

        self.conelength = sqrt((r1 - r0) ** 2 + self.length ** 2)
        self.side1 = (r1 - r0) / self.conelength
        self.side2 = self.length / self.conelength

        cdef double rmax = max(r0, r1)

        self._xlo = min(x0 - rmax, x1 - rmax)
        self._xhi = max(x0 + rmax, x1 + rmax)
        self._ylo = min(y0 - rmax, y1 - rmax)
        self._yhi = max(y0 + rmax, y1 + rmax)
        self._zlo = min(z0 - rmax, z1 - rmax)
        self._zhi = max(z0 + rmax, z1 + rmax)
        self.neighbors = []
        self.cx, self.cy, self.cz = (x0 + x1) * 0.5, (y0 + y1) * 0.5, (z0 + z1) * 0.5
        self.h = self.length * .5
        self.clips = []
        self.neighbor_regions = []

    def set_clip(self, clips):
        self.clips = clips
    def get_clip(self):
        return self.clips

    cpdef within_core(self, double px, double py, double pz):
        cdef double nx, ny, nz, y
        nx, ny, nz = px - self.cx, py - self.cy, pz - self.cz
        y = abs(self.axisx * nx + self.axisy * ny + self.axisz * nz)
        return y < self.h


    def axis(self):
        if self.reversed:
            return -self.axisx, -self.axisy, -self.axisz
        else:
            return (self.axisx, self.axisy, self.axisz)

    def starting_points(self, xs, ys, zs):
        # TODO: this only works right if the entire object is inside the domain
        return sum([c.starting_points(xs, ys, zs) for c in self.clips],
                    [(bisect.bisect_left(xs, self.x0), bisect.bisect_left(ys, self.y0), bisect.bisect_left(zs, self.z0)),
                     (bisect.bisect_left(xs, self.x1), bisect.bisect_left(ys, self.y1), bisect.bisect_left(zs, self.z1))])

    cpdef double distance(self, double px, double py, double pz):
        # CTNG:frustumdist
        cdef double nx, ny, nz, y, yy, xx, ry, rx, d
        nx, ny, nz = px - self.x0, py - self.y0, pz - self.z0
        y = nx * self.axisx + ny * self.axisy + nz * self.axisz
        yy = y * y
        xx = nx * nx + ny * ny + nz * nz - yy
        # in principle, xx >= 0, however roundoff errors may cause trouble
        if xx < 0: xx = 0

        if y < 0:
            # always nonnegative distance in this case
            if xx < self.rra:
                d = -y
            else:
                x = sqrt(xx) - self.r0
                d = sqrt(x * x + yy)
        elif xx < self.rrb and y > self.length:
            d = y - self.length
        else:
            x = sqrt(xx) - self.r0
            # y >= 0 always at this point (and if outside, not in the cylinder extending through the small end face)
            ry = x * self.side1 + y * self.side2
            if ry < 0:
                # if ry < 0 (and y > 0 from above), then outside the cone
                d = sqrt(x * x + yy)
            else:
                rx = x * self.side2 - y * self.side1
                if ry > self.conelength and y > self.length:
                    ry -= self.conelength
                    d = sqrt(rx * rx + ry * ry)
                else:
                    d = rx
                    if d < 0:
                        # end faces could be closer than the cone itself
                        d = max(rx, y - self.length)



        for clip in self.clips:
            d = max(d, clip.distance(px, py, pz))
        return d



    cpdef bint overlaps_x(self, double lo, double hi):
        return lo <= self._xhi and hi >= self._xlo
    cpdef bint overlaps_y(self, double lo, double hi):
        return lo <= self._yhi and hi >= self._ylo
    cpdef bint overlaps_z(self, double lo, double hi):
        return lo <= self._zhi and hi >= self._zlo


cdef class SkewCone:
    cdef double x0, y0, z0, r0, x1, y1, z1, r1, rra, rrb, axisx, axisy, axisz, conelength, side1, side2
    cdef double length, _xlo, _xhi, _ylo, _yhi, _zlo, _zhi, sx, sy, sz, planed
    property _x0:
        def __get__(self): return self.x0
    property _y0:
        def __get__(self): return self.y0
    property _z0:
        def __get__(self): return self.z0

    property xlo:
        def __get__(self): return self._xlo
    property xhi:
        def __get__(self): return self._xhi
    property ylo:
        def __get__(self): return self._ylo
    property yhi:
        def __get__(self): return self._yhi
    property zlo:
        def __get__(self): return self._zlo
    property zhi:
        def __get__(self): return self._zhi
    property primitives:
        def __get__(self): return [self]

    def get_clip(self):
        # TODO: change this if ever expand to allow clipping
        return []


    def __init__(self, double x0, double y0, double z0, double r0, double x1, double y1, double z1, double r1, double x2, double y2, double z2):
        """(x2, y2, z2) denotes point to skew (x1, y1, z1) to"""


        if r1 > r0:
            x0, y0, z0, r0, x1, y1, z1, r1, x2, y2, z2 = x2, y2, z2, r1, x2 - (x1 - x0), y2 - (y1 - y0), z2 - (z1 - z0), r0, x0, y0, z0

        self.x0, self.y0, self.z0, self.r0, self.x1, self.y1, self.z1, self.r1 = x0, y0, z0, r0, x1, y1, z1, r1
        self.rra, self.rrb = r0 * r0, r1 * r1
        self.axisx, self.axisy, self.axisz = (x1 - x0, y1 - y0, z1 - z0)

        self.length = sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2 + (z1 - z0) ** 2)
        # skew data
        self.sx = (x2 - x1) / self.length
        self.sy = (y2 - y1) / self.length
        self.sz = (z2 - z1) / self.length

        # normalize the axis
        self.axisx /= self.length
        self.axisy /= self.length
        self.axisz /= self.length

        self.planed = - (self.axisx * x0 + self.axisy * y0 + self.axisz * z0)

        self.conelength = sqrt((r1 - r0) ** 2 + self.length ** 2)
        self.side1 = (r1 - r0) / self.conelength
        self.side2 = self.length / self.conelength
        cdef double rmax = max(r0, r1)

        self._xlo = min(x0 - rmax, x2 - rmax)
        self._xhi = max(x0 + rmax, x2 + rmax)
        self._ylo = min(y0 - rmax, y2 - rmax)
        self._yhi = max(y0 + rmax, y2 + rmax)
        self._zlo = min(z0 - rmax, z2 - rmax)
        self._zhi = max(z0 + rmax, z2 + rmax)



    def starting_points(self, xs, ys, zs):
        # TODO: this only works right if the entire object is inside the domain
        return [(bisect.bisect_left(xs, self.x0), bisect.bisect_left(ys, self.y0), bisect.bisect_left(zs, self.z0))]
        #return sum([c.starting_points(xs, ys, zs) for c in self.clips], [(bisect.bisect_left(xs, self._x0), bisect.bisect_left(ys, self.y0), bisect.bisect_left(zs, self.z0))])

    # TODO: allow clipping?
    cpdef double distance(self, double px, double py, double pz):
        # CTNG:shearfrustdist
        cdef double nx, ny, nz, y, yy, xx, ry, rx, dist

        # compute signed distance of point to plane (note: axis vector has length 1)
        dist = px * self.axisx + py * self.axisy + pz * self.axisz + self.planed

        # deskew
        px -= dist * self.sx
        py -= dist * self.sy
        pz -= dist * self.sz

        nx, ny, nz = px - self.x0, py - self.y0, pz - self.z0
        y = nx * self.axisx + ny * self.axisy + nz * self.axisz
        yy = y * y
        xx = nx * nx + ny * ny + nz * nz - yy
        # in principle, xx >= 0, however roundoff errors may cause trouble
        if xx < 0: xx = 0

        if y < 0:
            if xx < self.rra:
                return -y
            else:
                x = sqrt(xx) - self.r0
                return sqrt(x * x + yy)
        elif xx < self.rrb and y > self.length:
            return y - self.length
        else:
            x = sqrt(xx) - self.r0
            ry = x * self.side1 + y * self.side2
            if ry < 0:
                return sqrt(x * x + yy)
            else:
                rx = x * self.side2 - y * self.side1
                if ry > self.conelength and y > self.length:
                    ry -= self.conelength
                    return sqrt(rx * rx + ry * ry)
                else:
                    if rx >= 0:
                        return rx
                    else:
                        return max(rx, y - self.length)



    cpdef bint overlaps_x(self, double lo, double hi):
        return lo <= self._xhi and hi >= self._xlo
    cpdef bint overlaps_y(self, double lo, double hi):
        return lo <= self._yhi and hi >= self._ylo
    cpdef bint overlaps_z(self, double lo, double hi):
        return lo <= self._zhi and hi >= self._zlo



# Avogadro's number (approximation before 2019 redefinition)
NA_legacy = 6.02214129e23
NA_modern = 6.02214076e23


def NA():
    try:
        from neuron import h

        # val = NA_legacy if h.nrnunit_use_legacy() else NA_modern
        val = h.Avogadro_constant
        # Note: h.Avogadro_constant is consistent with the above NA legacy and
        # modern values.
    except:
        val = NA_modern
    return val


def molecules_per_mM_um3():
    # converting from mM um^3 to molecules
    # = 6.02214129e23 * 1000. / 1.e18 / 1000
    # = avogadro * (L / m^3) * (m^3 / um^3) * (mM / M)
    return NA() / 1e18



# Basically want to test that net_move statement doesn't get
# mixed up with other instances.
import distutils.util
import os

from neuron import h

h.load_file("stdrun.hoc")

pc = h.ParallelContext()
h.steps_per_ms = 8
h.dt = 1.0 / h.steps_per_ms


class Cell:
    def __init__(self, gid):
        self.soma = h.Section(name="soma", cell=self)
        if gid % 2 == 0:
            # CoreNEURON permutation not the identity if cell topology not homogeneous
            self.dend = h.Section(name="dend", cell=self)
            self.dend.connect(self.soma(0.5))
        self.gid = gid
        pc.set_gid2node(gid, pc.id())
        self.r = h.Random()
        self.r.Random123(gid, 0, 0)
        self.syn = h.Bounce(self.soma(0.5))
        pc.cell(gid, h.NetCon(self.soma(0.5)._ref_v, None, sec=self.soma))
        self.syn.noiseFromRandom123(gid, 0, 1)
        self.t1vec = h.Vector()
        self.t1vec.record(self.syn._ref_t1, sec=self.soma)
        self.xvec = h.Vector()
        self.xvec.record(self.syn._ref_x, sec=self.soma)
        self.rvec = h.Vector()
        self.rvec.record(self.syn._ref_r, sec=self.soma)

    def result(self):
        return (
            self.syn.n_high,
            self.syn.n_mid,
            self.syn.n_low,
            self.t1vec.c(),
            self.xvec.c(),
            self.rvec.c(),
        )


def test_watchrange():
    from neuron import coreneuron

    coreneuron.enable = False

    ncell = 10
    gids = range(pc.id(), ncell, pc.nhost())  # round robin

    cells = [Cell(gid) for gid in gids]

    # complete the coverage of netcvode.cpp static void steer_val
    # Just so happens that Bounce declares an x var that does not get
    # mirrored by NetCon.x
    nc = h.NetCon(cells[2].syn, None)
    cells[2].syn.x = 0.1
    nc.x = 2.0
    assert nc.x == 0.0
    assert cells[2].syn.x == 0.1
    del nc

    # @olupton changed from 20 to trigger assert(datum==2) failure.
    tstop = 1.0

    def run(tstop, mode):
        pc.set_maxstep(10)
        h.finitialize(-65)
        if mode == 0:
            pc.psolve(tstop)
        elif mode == 1:
            while h.t < tstop:
                pc.psolve(h.t + h.dt)
        else:
            while h.t < tstop:
                h.continuerun(h.t + h.dt)
                pc.psolve(h.t + h.dt)

    tvec = h.Vector()
    tvec.record(h._ref_t, sec=cells[0].soma)
    run(tstop, 0)  # NEURON run
    tvec = tvec.c()  # don't record again but save.

    stdlist = [cell.result() for cell in cells]

    print("CoreNEURON run")
    h.CVode().cache_efficient(1)
    coreneuron.enable = True
    coreneuron.verbose = 0
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )

    def runassert(mode):
        run(tstop, mode)
        hml = ["invalid", "low", " mid", " high"]
        for i, cell in enumerate(cells):
            result = cell.result()
            std = stdlist[i]
            # Organised this way so we get a better overview of what went wrong
            # when something fails.
            success = all(std[j] == result[j] for j in range(3)) and all(
                std[j].eq(result[j]) for j in range(3, 6)
            )
            if not success:
                print("mode=" + str(mode))
                for j in range(3):
                    if std[j] != result[j]:
                        print(
                            "cell=%d %s:(nrn: %d cnrn: %d)"
                            % (i, ("high", "mid", "low")[j], std[j], result[j])
                        )
                # Look at the first place the flag value differs
                k = int(std[4].c().sub(result[4]).indwhere("!=", 0))
                print(
                    "first difference at %d (%g, %s, r=%g) vs (%g, %s, r=%g)"
                    % (
                        k,
                        std[3][k],
                        hml[int(std[4][k])],
                        std[5][k],
                        result[3][k],
                        hml[int(result[4][k])],
                        result[5][k],
                    )
                )
                for ik in range(k + 1):
                    print(
                        "  %d %g nrn(%g, %s, r=%g) vs cnrn(%g, %s, r=%g)"
                        % (
                            ik,
                            tvec[ik],
                            std[3][ik],
                            hml[int(std[4][ik])],
                            std[5][ik],
                            result[3][ik],
                            hml[int(result[4][ik])],
                            result[5][ik],
                        )
                    )

            assert success

    for mode in [0, 1, 2]:
        runassert(mode)

    coreneuron.enable = False
    # teardown
    pc.gid_clear()
    return stdlist, tvec


if __name__ == "__main__":
    from neuron import gui

    stdlist, tvec = test_watchrange()
    g = h.Graph()
    print("n_high  n_mid  n_low")
    for i, result in enumerate(stdlist):
        print(result[0], result[1], result[2])
        result[4].line(g, tvec, i, 2)
    g.exec_menu("View = plot")
    h.quit()



def test(files, correct_data):
    import os
    import sys
    import filecmp
    import subprocess
    import re, array, numpy

    tol = 1e-10
    dt_eps = 1e-20
    for dr in ["wave1d", "ecs", "3d", "hybrid"]:
        try:
            os.makedirs(os.path.join("test_data", dr))
        except OSError:
            pass

    passed_list = []
    failed_list = []
    incomplete_list = []

    for f in files:
        base_name = f[:-3]
        print("%s: " % base_name)

        output_file = os.path.join("test_data", base_name + ".dat")
        # remove any old test results
        try:
            os.remove(output_file)
        except OSError:
            pass

        #        os.system('python do_test.py %s %s' % (os.path.join('tests', f), output_file))

        try:
            outp = subprocess.check_output(
                [sys.executable, "do_test.py", os.path.join("tests", f), output_file]
            )
            sobj = re.search(r"<BAS_RL (\d*) BAS_RL>", outp.decode("utf-8"), re.M)
            rlen = int(sobj.group(1))
            success = False
            corr_dat = numpy.fromfile(
                os.path.join(correct_data, base_name + ".dat")
            ).reshape(-1, rlen)
            tst_dat = numpy.fromfile(output_file).reshape(-1, rlen)
            t1 = corr_dat[:, 0]
            t2 = tst_dat[:, 0]
            # remove any initial t that are greter than the next t (removes times before 0) in correct data
            c = 0
            while c < len(t1) - 1 and t1[c] > t1[c + 1]:
                c = c + 1
            t1 = numpy.delete(t1, range(c))
            corr_dat = numpy.delete(corr_dat, range(c), 0)
            # remove any initial t that are greter than the next t (removes times before 0) in test data
            c = 0
            while c < len(t2) - 1 and t2[c] > t2[c + 1]:
                c = c + 1
            t2 = numpy.delete(t2, range(c))
            tst_dat = numpy.delete(tst_dat, range(c), 0)
            # get rid of repeating t in correct data (otherwise interpolation fails)
            c = 0
            while c < len(t1) - 1:
                c1 = c + 1
                while c1 < len(t1) and abs(t1[c] - t1[c1]) < dt_eps:
                    c1 = c1 + 1
                t1 = numpy.delete(t1, range(c, c1 - 1))
                corr_dat = numpy.delete(corr_dat, range(c, c1 - 1), 0)
                c = c + 1
            # get rid of the test data outside of the correct data time interval
            t2_n = len(t2)
            t2_0 = 0
            while t2[t2_n - 1] > t1[-1]:
                t2_n = t2_n - 1
            while t2[t2_0] < t1[0]:
                t2_0 = t2_0 + 1
            # interpolate and compare
            corr_vals = numpy.array(
                [
                    numpy.interp(t2[t2_0:t2_n], t1, corr_dat[:, i].T)
                    for i in range(1, rlen)
                ]
            )
            max_err = numpy.amax(abs(corr_vals.T - tst_dat[t2_0:t2_n, 1:]))

            if max_err < tol:
                success = True

            if success:
                passed_list.append(base_name)
                print("passed")
            else:
                print("failed")

                failed_list.append(base_name)
        except Exception as e:
            print("incomplete", e)
            incomplete_list.append(base_name)

    print("")
    print("---")
    print("Passed:     ", ", ".join(passed_list))
    print("Failed:     ", ", ".join(failed_list))
    print("Incomplete: ", ", ".join(incomplete_list))
    print("---")
    print("Ran %d tests:" % len(files))
    print("Passed:     %d" % len(passed_list))
    print("Failed:     %d" % len(failed_list))
    print("Incomplete: %d" % len(incomplete_list))
    if len(passed_list) < len(files):
        return -1
    return 0


if __name__ == "__main__":
    import os
    import sys
    import shutil
    import subprocess
    import platform

    abspath = os.path.abspath(__file__)
    dname = os.path.dirname(abspath)
    os.chdir(dname)

    rxd_data = os.path.join(
        os.pardir, os.pardir, os.pardir, os.pardir, os.pardir, "test", "rxd", "testdata"
    )
    # get the rxd test data
    subprocess.call(["git", "submodule", "update", "--init", "--recursive", rxd_data])
    correct_data = os.path.abspath(os.path.join(rxd_data, "rxdtests"))

    files = [f for f in os.listdir("tests") if f[-3:].lower() == ".py"]
    # dirs = [name for name in os.listdir('tests') if os.path.isdir(os.path.join('tests', name))]
    dirs = [
        os.path.join(*d.split(os.path.sep)[1:])
        for d in [wlk[0] for wlk in os.walk("tests")][1:]
    ]
    os.chdir(dname)

    for dr in dirs:
        fname = os.path.join("tests", dr, "torun.txt")
        if os.path.isfile(fname):
            # compile any mod files
            os.chdir(os.path.join("tests", dr))
            # remove old compiled files
            try:
                shutil.rmtree(platform.machine())
            except OSError:
                pass
            os.system("nrnivmodl")
            os.chdir(dname)
            with open(fname) as f:
                files += [os.path.join(dr, s) for s in f.read().splitlines()]
    sys.exit(test(files, correct_data))



# -*- coding: utf-8 -*-
import warnings
import numpy
from neuron import h, nrn
from .rxdException import RxDException

try:
    from neuron.rxd import geometry3d

    has_geometry3d = True
except ImportError:
    has_geometry3d = False


class RxDGeometry:
    def volumes1d(self, sec):
        raise RxDException("volume1d unimplemented")

    def surface_areas1d(self, sec):
        raise RxDException("surface_areas1d unimplemented")

    def neighbor_areas1d(self, sec):
        raise RxDException("neighbor_areas1d unimplemented")

    def is_volume(self):
        raise RxDException("is_volume unimplemented")

    def is_area(self):
        raise RxDException("is_area unimplemented")

    def __call__(self):
        """calling returns self to allow for rxd.inside or rxd.inside()"""
        return self


def _volumes1d(sec):
    if not isinstance(sec, nrn.Section):
        sec = sec._sec
    arc3d = [sec.arc3d(i) for i in range(sec.n3d())]
    diam3d = [sec.diam3d(i) for i in range(sec.n3d())]
    vols = numpy.zeros(sec.nseg)
    dx = sec.L / sec.nseg
    for iseg in range(sec.nseg):
        # get a list of all pts in the segment, including end points
        lo = iseg * dx
        hi = (iseg + 1) * dx
        pts = [lo] + [x for x in arc3d if lo < x < hi] + [hi]

        diams = numpy.interp(pts, arc3d, diam3d)

        # sum the volume of the constituent frusta
        volume = 0
        for i in range(len(pts) - 1):
            diam0, diam1 = diams[i : i + 2]
            pt0, pt1 = pts[i : i + 2]
            volume += (
                numpy.pi
                * (pt1 - pt0)
                / 12.0
                * (diam0**2 + diam0 * diam1 + diam1**2)
            )
        vols[iseg] = volume

    return vols


def _make_surfacearea1d_function(scale, diam_scale=1.0):
    def result(sec):
        if not isinstance(sec, nrn.Section):
            sec = sec._sec
        arc3d = [sec.arc3d(i) for i in range(sec.n3d())]
        diam3d = [sec.diam3d(i) * diam_scale for i in range(sec.n3d())]
        sas = numpy.zeros(sec.nseg)
        dx = sec.L / sec.nseg
        for iseg in range(sec.nseg):
            # get a list of all pts in the segment, including end points
            lo = iseg * dx
            hi = (iseg + 1) * dx
            pts = [lo] + [x for x in arc3d if lo < x < hi] + [hi]

            diams = numpy.interp(pts, arc3d, diam3d)

            # sum the surface areas of the constituent frusta
            sa = 0
            for i in range(len(pts) - 1):
                diam0, diam1 = diams[i : i + 2]
                pt0, pt1 = pts[i : i + 2]
                sa += (
                    scale
                    * 0.5
                    * (diam0 + diam1)
                    * numpy.sqrt(0.25 * (diam0 - diam1) ** 2 + (pt1 - pt0) ** 2)
                )
            sas[iseg] = sa
        return sas

    return result


def _make_perimeter_function(scale, diam_scale=1.0):
    def result(sec):
        if not isinstance(sec, nrn.Section):
            sec = sec._sec
        arc3d = [sec.arc3d(i) for i in range(sec.n3d())]
        diam3d = [sec.diam3d(i) * diam_scale for i in range(sec.n3d())]
        area_pos = numpy.linspace(0, sec.L, sec.nseg + 1)
        diams = numpy.interp(area_pos, arc3d, diam3d)
        return scale * diams

    return result


_surface_areas1d = _make_surfacearea1d_function(numpy.pi)
_perimeter1d = _make_perimeter_function(numpy.pi)


def _neighbor_areas1d(sec):
    if not isinstance(sec, nrn.Section):
        sec = sec._sec
    arc3d = [sec.arc3d(i) for i in range(sec.n3d())]
    diam3d = [sec.diam3d(i) for i in range(sec.n3d())]
    area_pos = numpy.linspace(0, sec.L, sec.nseg + 1)
    diams = numpy.interp(area_pos, arc3d, diam3d)
    return numpy.pi * 0.25 * diams**2


def constant_function_per_length(value):
    return lambda sec: [value * sec.L / sec.nseg for i in range(sec.nseg)]


def constant_everywhere_1d(value):
    return lambda sec: value * numpy.ones(sec.nseg)


def constant_everywhere_plus_one_1d(value):
    return lambda sec: value * numpy.ones(sec.nseg + 1)


def constant_function(value):
    return lambda *args, **kwargs: value


def scale_by_constant(scale, f):
    return lambda *args, **kwargs: scale * f(*args, **kwargs)


_always_true = constant_function(True)
_always_false = constant_function(False)
_always_0 = constant_function(0)

inside = RxDGeometry()
if has_geometry3d:
    inside.volumes3d = geometry3d.voxelize2
# neighbor_area_fraction can be a constant or a function
inside.neighbor_area_fraction = 1
inside.volumes1d = _volumes1d
inside.surface_areas1d = _surface_areas1d
inside.neighbor_areas1d = _neighbor_areas1d
inside.is_volume = _always_true
inside.is_area = _always_false
inside.__repr__ = constant_function("inside")

# TODO: make a version that allows arbitrary shells?
membrane = RxDGeometry()
membrane.volumes1d = _surface_areas1d
membrane.surface_areas1d = _always_0
membrane.neighbor_areas1d = _perimeter1d
membrane.is_volume = _always_false
membrane.is_area = _always_true
membrane.__repr__ = constant_function("membrane")


class Enum:
    """a silly way of creating unique identifiers without using/allowing/requiring magic constants"""

    pass


_lo_hi_shell = Enum()


class DistributedBoundary(RxDGeometry):
    """Boundary that scales with area.

    DistributedBoundary(area_per_vol, perim_per_area=0)

    area_per_vol is the area of the boundary as a function of the volume
    containing it. e.g.

    g = DistributedBoundary(2) defines a geometry with 2 um^2 of area per
    every um^3 of volume.

    perim_per_area is the perimeter (in um) per 1 um^2 cross section of the
    volume. For use in reaction-diffusion problems, it may be safely omitted
    if and only if no species in the corresponding region diffuses.

    This is often useful for separating FractionalVolume objects.

    It is assumed that the area is always strictly on the interior.
    """

    def __init__(self, area_per_vol, perim_per_area=0):
        self._area_per_vol = area_per_vol
        self._perim_per_area = 0

        self.surface_areas1d = _always_0
        self.neighbor_areas1d = scale_by_constant(perim_per_area, _neighbor_areas1d)
        self.volumes1d = scale_by_constant(area_per_vol, _volumes1d)
        self.is_volume = _always_false
        self.is_area = _always_true

    @property
    def neighbor_area_fraction(self):
        # TODO: validate that this gives consistent results between 1D and 3D
        return self._perim_per_area

    def volumes3d(
        self,
        source,
        dx=0.25,
        xlo=None,
        xhi=None,
        ylo=None,
        yhi=None,
        zlo=None,
        zhi=None,
        n_soma_step=100,
    ):
        # mesh, surface_areas, volumes, triangles = geometry3d.voxelize2(source, dx=dx)
        # volumes._values *= self._area_per_vol # volume on 2D boundaries is actually the area; the amount of space for holding things
        # surface_areas._values *= 0
        # return mesh, surface_areas, volumes, triangles
        internal_voxels, surface_voxels, mesh_grid = geometry3d.voxelize2(source, dx=dx)
        area_per_vol = self._area_per_vol
        for key in internal_voxels:
            internal_voxels[key][0] *= area_per_vol
        for key in surface_voxels:
            surface_voxels[key][0] *= area_per_vol
        return internal_voxels, surface_voxels, mesh_grid

    def __repr__(self):
        if self._perim_per_area == 0:
            return "DistributedBoundary(%g)" % (self._area_per_vol)
        else:
            return "DistributedBoundary(%g, perim_per_area=%g)" % (
                self._area_per_vol,
                self._perim_per_area,
            )


class FractionalVolume(RxDGeometry):
    def __init__(
        self, volume_fraction=1, surface_fraction=0, neighbor_areas_fraction=None
    ):
        if neighbor_areas_fraction is None:
            neighbor_areas_fraction = volume_fraction
        if surface_fraction == 0:
            self.surface_areas1d = _always_0
        elif surface_fraction == 1:
            self.surface_areas1d = _surface_areas1d
        else:
            self.surface_areas1d = scale_by_constant(surface_fraction, _surface_areas1d)
        # TODO: add the if statement so not scaling if 0 or 1
        self.neighbor_areas1d = scale_by_constant(
            neighbor_areas_fraction, _neighbor_areas1d
        )
        self.volumes1d = scale_by_constant(volume_fraction, _volumes1d)
        self.is_volume = _always_true
        self.is_area = _always_false
        self._volume_fraction = volume_fraction
        self._surface_fraction = surface_fraction
        self._neighbor_areas_fraction = neighbor_areas_fraction
        # TODO: does the else case ever make sense?
        self.neighbor_area_fraction = (
            volume_fraction
            if neighbor_areas_fraction is None
            else neighbor_areas_fraction
        )

    def volumes3d(
        self,
        source,
        dx=0.25,
        xlo=None,
        xhi=None,
        ylo=None,
        yhi=None,
        zlo=None,
        zhi=None,
        n_soma_step=100,
    ):
        # mesh, surface_areas, volumes, triangles = geometry3d.voxelize2(source, dx=dx)
        # surface_areas._values *= self._surface_fraction
        # volumes._values *= self._volume_fraction
        # return mesh, surface_areas, volumes, triangles
        internal_voxels, surface_voxels, mesh_grid = geometry3d.voxelize2(source, dx=dx)
        volume_fraction = self._volume_fraction
        for key in internal_voxels:
            internal_voxels[key][0] *= volume_fraction
        for key in surface_voxels:
            surface_voxels[key][0] *= volume_fraction
        return internal_voxels, surface_voxels, mesh_grid

    def __repr__(self):
        return (
            "FractionalVolume(volume_fraction=%r, surface_fraction=%r, neighbor_areas_fraction=%r)"
            % (
                self._volume_fraction,
                self._surface_fraction,
                self._neighbor_areas_fraction,
            )
        )


# TODO: eliminate this class and replace with FixedCrossSection?
class ConstantVolume(RxDGeometry):
    # TODO: do we want different default neighbor_area?
    def __init__(self, volume=1, surface_area=0, neighbor_area=1):
        """volume, surface_area per unit length"""
        self.volumes1d = constant_function_per_length(volume)
        self.surface_areas1d = constant_function_per_length(surface_area)
        self.is_volume = _always_true
        self.is_area = _always_false
        self.neighbor_areas1d = constant_everywhere_plus_one_1d(neighbor_area)


class FixedCrossSection(RxDGeometry):
    def __init__(self, cross_area, surface_area=0):
        self.volumes1d = constant_function_per_length(cross_area)
        self.surface_areas1d = constant_function_per_length(surface_area)
        self.is_volume = _always_true
        self.is_area = _always_false
        self.neighbor_areas1d = constant_everywhere_plus_one_1d(cross_area)
        self._cross_area = cross_area
        self._surface_area = surface_area

    def __repr__(self):
        return "FixedCrossSection(%r, surface_area=%r)" % (
            self._cross_area,
            self._surface_area,
        )


class FixedPerimeter(RxDGeometry):
    def __init__(self, perimeter, on_cell_surface=False):
        self.volumes1d = constant_function_per_length(perimeter)
        self.surface_areas1d = _always_0 if not on_cell_surface else self.volumes1d
        self._perim = perimeter
        self.is_volume = _always_false
        self.is_area = _always_true
        self._on_surface = on_cell_surface

    def neighbor_areas1d(self, sec):
        return [self._perim] * (sec.nseg + 1)

    def __repr__(self):
        return "FixedPerimeter(%r, on_cell_surface=%r)" % (
            self._perim,
            self._on_surface,
        )


class ScalableBorder(RxDGeometry):
    """a membrane that scales proportionally with the diameter.

    Example use:

    - the boundary between radial shells e.g.
      ScalableBorder(diam_scale=0.5) could represent the border of
      Shell(lo=0, hi=0.5)


    Args:
        scale (float, optional) scale the area, default value is π.
            e.g. for a cylinder of length L and diameter d, ScalableBorder will
            give an area scale*d*L, by default the surface area.
            For cylindrical sections only. Use "diam_scale" instead to correctly
            handle cylindrical and non-cylindrical sections.
        diam_scale (float, optional), scale the diameter, default value is 1.
            e.g. for a cylinder of length L and diameter d, ScalableBorder will
            give an area diam_scale*π*d*L, by default the surface area.

    Note: Provide either a scale or diam_scale, not both.

    Sometimes useful for the boundary between FractionalVolume objects, but
    see also DistributedBoundary which scales with area.
    """

    def __init__(self, scale=None, diam_scale=None, on_cell_surface=False):
        if scale is not None and diam_scale is not None:
            raise RxDException(
                "ScalableBorder either provide scale or diam_scale, not both"
            )
        elif diam_scale is not None:
            self._scale = numpy.pi
            self._diam_scale = diam_scale
        elif scale is not None:
            self._scale = scale
            self._diam_scale = 1.0
        else:
            self._scale = numpy.pi
            self._diam_scale = 1.0

        self.volumes1d = _make_surfacearea1d_function(self._scale, self._diam_scale)
        self.surface_areas1d = _always_0 if not on_cell_surface else self.volumes1d
        self.is_volume = _always_false
        self.is_area = _always_true
        self.neighbor_areas1d = _make_perimeter_function(self._scale, self._diam_scale)
        self._on_surface = on_cell_surface

    def __repr__(self):
        return "ScalableBorder(%r, on_cell_surface=%r)" % (
            self._scale,
            self._on_surface,
        )


# TODO: remove this, use FixedPerimeter instead?
class ConstantArea(RxDGeometry):
    def __init__(self, area=1, perim=1, on_cell_surface=False):
        # TODO: fix this
        warnings.warn("ConstantArea probably not right")
        self.volumes1d = constant_function(area)
        self.surface_areas1d = (
            _always_0 if not on_cell_surface else constant_function(area)
        )
        self._perim = perim
        self.is_volume = _always_false
        self.is_area = _always_true

    def neighbor_areas1d(self, sec):
        return [self._perim] * (sec.nseg + 1)


# TODO: is there a better name than Shell?
class Shell(RxDGeometry):
    def __init__(self, lo=None, hi=None):
        if lo is None or hi is None:
            raise RxDException("only Shells with a lo and hi are supported for now")

        if lo > hi:
            lo, hi = hi, lo
        if lo == hi:
            raise RxDException("Shell objects must have thickness")
        self._type = _lo_hi_shell
        self._lo = lo
        self._hi = hi

        if lo == 1 or hi == 1:
            self.surface_areas1d = _surface_areas1d
        elif lo < 1 < hi:
            raise RxDException(
                "shells may not cross the membrane; i.e. 1 cannot lie strictly between lo and hi"
            )
        else:
            # TODO: is this what we want; e.g. what if lo < 1 < hi?
            self.surface_areas1d = _always_0

    def __repr__(self):
        return "Shell(lo=%r, hi=%r)" % (self._lo, self._hi)

    def neighbor_areas1d(self, sec):
        if not isinstance(sec, nrn.Section):
            sec = sec._sec
        arc3d = [sec.arc3d(i) for i in range(sec.n3d())]
        diam3d = [sec.diam3d(i) for i in range(sec.n3d())]
        area_pos = numpy.linspace(0, sec.L, sec.nseg + 1)
        diams = numpy.interp(area_pos, arc3d, diam3d)
        if self._type == _lo_hi_shell:
            return numpy.pi * 0.25 * ((diams * self._hi) ** 2 - (diams * self._lo) ** 2)

    def is_volume(self):
        return True

    def is_area(self):
        return False

    def volumes1d(self, sec):
        if not isinstance(sec, nrn.Section):
            sec = sec._sec
        arc3d = [sec.arc3d(i) for i in range(sec.n3d())]
        diam3d = [sec.diam3d(i) for i in range(sec.n3d())]
        vols = numpy.zeros(sec.nseg)
        dx = sec.L / sec.nseg
        for iseg in range(sec.nseg):
            # get a list of all pts in the segment, including end points
            lo = iseg * dx
            hi = (iseg + 1) * dx
            pts = [lo] + [x for x in arc3d if lo < x < hi] + [hi]

            diams = numpy.interp(pts, arc3d, diam3d)

            # sum the volume of the constituent frusta, hollowing out by the inside
            volume = 0
            for i in range(len(pts) - 1):
                diam0h, diam1h = self._hi * diams[i : i + 2]
                diam0l, diam1l = self._lo * diams[i : i + 2]
                pt0, pt1 = pts[i : i + 2]
                volume += (
                    numpy.pi
                    * (pt1 - pt0)
                    / 12.0
                    * (
                        (diam0h**2 + diam0h * diam1h + diam1h**2)
                        - (diam0l**2 + diam0l * diam1l + diam1l**2)
                    )
                )
            vols[iseg] = volume

        return vols


class MultipleGeometry(RxDGeometry):
    """support for different geometries on different sections of a region.

    Example use:
    - for radial diffusion in a dendrite (dend) with longitudinal diffusion
      from a spine (spine). The region for the outer shell of the dendrite
      (0.8,1] should include the while spine [0,1];
      MultipleGeometry(secs=[dend,spine], geos=[Shell(0.8,1), rxd.inside])

    Args:
    sections (list, optional) a list or list-of-lists of sections where the
        corresponding geometry should be used. If None the same geometry used
        for all sections, otherwise the list must be the same length as the
        geos list.
        If None is in the list, then the corresponding geometry in geos is used
        as a default value for any section not included in the lists.
    geometries (list) a list of geometries that are used for the corresponding
        list of sections in secs. All geometries must be volumes or all
        all geometries must be areas.

    """

    def __init__(self, secs=None, geos=None):
        self._secs = {}
        self._default = None
        if not secs:
            if isinstance(geos, list):
                self._default = geos[0]
            elif isinstance(geos, RxDGeometry):
                self._default = geos
            else:
                raise RxDException(
                    "MultipleGeometry requires a list-of-lists of sections and their corresponding geometry"
                )
        else:
            assert len(secs) == len(geos)
            if all([g.is_area() for g in geos]):
                self.is_area = _always_true
                self.is_volume = _always_false
            elif all([g.is_volume() for g in geos]):
                self.is_area = _always_false
                self.is_volume = _always_true
            else:
                raise RxDException(
                    "MultipleGeometry requires all geometries are areas or all geometries are volumes"
                )
            for s, g in zip(secs, geos):
                if not s:
                    self._default = g
                elif isinstance(s, list):
                    self._secs[h.SectionList(s)] = g
                else:
                    self._secs[h.SectionList([s])] = g

    def __repr__(self):
        secs = [[s for s in sl] for sl in self._secs]
        geos = [self._secs[sl] for sl in self._secs]
        return "MultipleGeometry(secs=%r, geos=%r)" % (secs, geos)

    def _get_geo(self, sec):
        if not isinstance(sec, nrn.Section):
            sec = sec._sec
        for sl in self._secs:
            if sec in sl:
                geo = self._secs[sl]
                break
        else:
            if self._default:
                geo = self._default
            else:
                raise RxDException(
                    "MultipleGeometry is not defined on section %r" % sec
                )
        return geo

    def volumes1d(self, sec):
        return self._get_geo(sec).volumes1d(sec)

    def surface_areas1d(self, sec):
        return self._get_geo(sec).surface_areas1d(sec)

    def neighbor_areas1d(self, sec):
        return self._get_geo(sec).neighbor_areas1d(sec)



# arrange for InputPreSyn to be on queue after finitialize

from neuron import h

pc = h.ParallelContext()


def sortspikes(spiketime, gidvec):
    return sorted(zip(spiketime, gidvec))


def test_inputpresyn():
    # NetStim with gid = 1 connected to IntFire1 with gid = 2
    # sadly IntFire1 does not exist in coreneuron so use IntervalFire
    # make cells
    nstim = 3
    ncell = 15
    cells = {gid: None for gid in range(pc.id(), ncell, pc.nhost())}
    for gid in cells:
        pc.set_gid2node(gid, pc.id())
        cells[gid] = h.NetStim() if gid < nstim else h.IntervalFire()
        pc.cell(gid, h.NetCon(cells[gid], None))

    # connect
    netcons = {}
    for gid in cells:
        if gid >= nstim:
            for srcgid in range(nstim):
                netcons[(srcgid, gid)] = nc = pc.gid_connect(srcgid, cells[gid])
                nc.delay = 1
                nc.weight[0] = 2
        else:  # The NetStim
            ns = cells[gid]
            ns.start = 0
            ns.number = 1
            ns.interval = 10

    # does it look like what we want?
    if False:
        for gid in cells:
            print(pc.id(), gid, cells[gid])
        for con in netcons:
            print(pc.id(), con)

    spiketime = h.Vector()
    gidvec = h.Vector()
    pc.spike_record(-1, spiketime, gidvec)

    def run(tstop):
        spiketime.resize(0)
        gidvec.resize(0)
        pc.set_maxstep(10)
        h.finitialize()
        pc.psolve(tstop)

    run(2)
    spikes_std = sortspikes(spiketime, gidvec)

    def same():
        spikes = sortspikes(spiketime, gidvec)
        assert spikes_std == spikes

    h.CVode().cache_efficient(1)
    from neuron import coreneuron

    coreneuron.enable = 1
    coreneuron.verbose = 0
    run(2)
    same()

    pc.barrier()


if __name__ == "__main__":
    test_inputpresyn()

    if pc.nhost() > 1:
        pc.barrier()
        h.quit()



from neuron import h


def test_random_play():  # for coverage of ptrlist changes #1815
    cv = h.CVode()
    cv.active(0)
    cv.cache_efficient(0)
    h.secondorder = 0
    s = h.Section()
    s.L = 10
    s.diam = 10
    ic = h.IClamp(s(0.5))
    ic.delay = 0
    ic.dur = 10
    ic.amp = 0
    r = h.Random()
    r.Random123(1, 0, 0)
    r.uniform(0.0, 0.001)
    r.play(ic._ref_amp)
    v_vec = h.Vector()
    v_vec.record(s(0.5)._ref_v, sec=s)
    h.dt = 0.025
    h.finitialize(0)
    for _ in range(3):
        h.fadvance()
    std = h.Vector(
        [0.0, 0.00046940111168614074, 0.004571699024450762, 0.01086603263672303]
    )
    assert v_vec.c().sub(std).abs().sum() < 1e-15


def test_hoc_list():  # for coverage of ptrlist changes in ivoc/oc_list.cpp
    lst = h.List()
    lst.append(h.Vector().append(0))
    lst.append(h.Vector().append(1))
    lst.insrt(1, h.Vector().append(2))
    assert [v.x[0] for v in lst] == [0.0, 2.0, 1.0]


def test_rvp():  # for coverage of ptrlist changes in nrniv/shape.cpp
    d = [h.Section(name="d_" + str(i)) for i in range(3)]
    for i in range(1, 3):
        d[i].connect(d[i - 1](1))
    for s in d:
        s.L = 10
        s.diam = 1
        s.insert("pas")
    rvp = h.RangeVarPlot("v", d[0](0), d[2](1))
    assert rvp.left() == 0.0
    assert rvp.right() == 30.0
    sl = h.SectionList()
    rvp.list(sl)
    sz = sum([1 for _ in sl])
    assert sz == 3


def test_finithnd():
    # ptrlist #1815 coverage of nrniv/finithnd.cpp
    import sys, io

    def foo():
        pass

    fih = h.FInitializeHandler(foo)
    oldstdout = sys.stdout
    sys.stdout = mystdout = io.StringIO()
    fih.allprint()
    sys.stdout = oldstdout
    assert "Type 1 FInitializeHandler statements" in mystdout.getvalue()


if __name__ == "__main__":
    test_random_play()
    test_hoc_list()
    test_rvp()
    test_finithnd()



# ParallelContext transfer functionality tests.
# This uses nonsense models and values to verify transfer takes place correctly.

import sys
from io import StringIO

from neuron import h

# h.nrnmpi_init()

pc = h.ParallelContext()
rank = pc.id()
nhost = pc.nhost()

if nhost > 1:
    if rank == 0:
        print("nhost > 1 so calls to expect_error will return without testing.")


def expect_error(callable, args, sec=None):
    """
    Execute callable(args) and assert that it generated an error.

    If sec is not None, executes callable(args, sec=sec)
    Skips if nhost > 1 as all hoc_execerror end in MPI_ABORT
    Does not work well with nrniv launch since hoc_execerror messages do not
    pass through sys.stderr.
    """

    if nhost > 1:
        return

    old_stderr = sys.stderr
    sys.stderr = my_stderr = StringIO()
    err = 0
    try:
        if sec:
            callable(*args, sec=sec)
        else:
            callable(*args)
    except:
        err = 1
    errmes = my_stderr.getvalue()
    sys.stderr = old_stderr
    if errmes:
        errmes = errmes.splitlines()[0]
        errmes = errmes[(errmes.find(":") + 2) :]
        print("expect_error: %s" % errmes)
    if err == 0:
        print("expect_error: no err for %s%s" % (str(callable), str(args)))
    assert err


# HGap POINT_PROCESS via ChannelBUilder.
#  Cannot use with extracellular.
ks = h.KSChan(1)
ks.name("HGap")
ks.iv_type(0)
ks.gmax(0)
ks.erev(0)

# Cell with enough nonsense stuff to exercise transfer possibilities.
class Cell:
    def __init__(self):
        self.soma = h.Section(name="soma", cell=self)
        self.soma.diam = 10.0
        self.soma.L = 10.0
        self.soma.insert("na_ion")  # can use nai as transfer source
        # can use POINT_PROCESS range variable as targets
        self.ic = h.IClamp(self.soma(0.5))
        self.vc = h.SEClamp(self.soma(0.5))
        self.vc.rs = 1e9  # no voltage clamp current
        self.hgap = [None for _ in range(2)]  # filled by mkgaps


def run():
    pc.setup_transfer()
    h.finitialize()
    h.fadvance()


model = None  # Global allows teardown of model


def teardown():
    """
    destroy model
    """
    global model
    pc.gid_clear()
    model = None


def mkmodel(ncell):
    """
    Destroy existing model and re-create with ncell Cells.
    """
    global model
    if model:
        teardown()
    cells = {}
    for gid in range(rank, ncell, nhost):
        cells[gid] = Cell()
        pc.set_gid2node(gid, rank)
        pc.cell(gid, h.NetCon(cells[gid].soma(0.5)._ref_v, None, sec=cells[gid].soma))
    model = (cells, ncell)


def mkgaps(gids):
    """For list of gids, full gap, right to left"""
    gidset = set()
    for gid in gids:
        g = [gid, (gid + 1) % model[1]]
        sids = [i + 1000 for i in g]
        for i, j in enumerate([1, 0]):
            if pc.gid_exists(g[i]):
                cell = model[0][g[i]]
                if g[i] not in gidset:  # source var sid cannot be used twice
                    pc.source_var(cell.soma(0.5)._ref_v, sids[i], sec=cell.soma)
                    gidset.add(g[i])
                assert cell.hgap[j] is None
                cell.hgap[j] = h.HGap(cell.soma(0.5))
                pc.target_var(cell.hgap[j], cell.hgap[j]._ref_e, sids[j])
                cell.hgap[j].gmax = 0.0001


def transfer1(amp1=True):
    """
    round robin transfer v to ic.amp and vc.amp1, nai to vc.amp2
    """
    ncell = model[1]
    for gid, cell in model[0].items():
        s = cell.soma
        srcsid = gid
        tarsid = (gid + 1) % ncell
        pc.source_var(s(0.5)._ref_v, srcsid, sec=s)
        pc.source_var(s(0.5)._ref_nai, srcsid + ncell, sec=s)
        pc.target_var(cell.ic, cell.ic._ref_amp, tarsid)
        if amp1:
            pc.target_var(cell.vc, cell.vc._ref_amp1, tarsid)
        pc.target_var(cell.vc, cell.vc._ref_amp2, tarsid + ncell)


def init_values():
    """
    Initialize sources to their sid values and targets to 0
    This allows substantive test that source values make it to targets.
    """
    ncell = model[1]
    for gid, c in model[0].items():
        c.soma(0.5).v = gid
        c.soma(0.5).nai = gid + ncell
        c.ic.amp = 0
        c.vc.amp1 = 0
        c.vc.amp2 = 0


def check_values():
    """
    Verify that target values are equal to source values.
    """
    values = {}
    for gid, c in model[0].items():
        vi = c.soma(0.5).v
        if h.ismembrane("extracellular", sec=c.soma):
            vi += c.soma(0.5).vext[0]
        values[gid] = {
            "v": vi,
            "nai": c.soma(0.5).nai,
            "amp": c.ic.amp,
            "amp1": c.vc.amp1,
            "amp2": c.vc.amp2,
        }
    x = pc.py_gather(values, 0)
    if rank == 0:
        values = {}
        for v in x:
            values.update(v)
        ncell = len(values)
        for gid in values:
            v1 = values[gid]
            v2 = values[(gid + ncell - 1) % ncell]
            assert v1["v"] == v2["amp"]
            assert v1["v"] == v2["amp1"]
            assert v1["nai"] == v2["amp2"]


def test_partrans():

    # no transfer targets or sources.
    mkmodel(4)
    run()

    # invalid source or target sid.
    if 0 in model[0]:
        cell = model[0][0]
        s = cell.soma
        expect_error(pc.source_var, (s(0.5)._ref_v, -1), sec=s)
        expect_error(pc.target_var, (cell.ic, cell.ic._ref_amp, -1))

    # target with no source.
    if pc.gid_exists(1):
        cell = pc.gid2cell(1)
        pc.target_var(cell.ic, cell.ic._ref_amp, 1)
    expect_error(run, ())

    mkmodel(4)

    # source with no target (not an error).
    if pc.gid_exists(1):
        cell = pc.gid2cell(1)
        pc.source_var(cell.soma(0.5)._ref_v, 1, sec=cell.soma)
    run()

    # No point process for target
    if pc.gid_exists(1):
        cell = pc.gid2cell(1)
        pc.target_var(cell.vc._ref_amp3, 1)
    try:
        run()  # ok if test_fast_imem.py not prior
    except:
        pass
    pc.nthread(2)
    expect_error(run, ())  # Do not know the POINT_PROCESS target
    pc.nthread(1)

    # Wrong sec for source ref and wrong point process for target ref.
    mkmodel(1)
    if pc.gid_exists(0):
        cell = pc.gid2cell(0)
        sec = h.Section(name="dend")
        expect_error(pc.source_var, (cell.soma(0.5)._ref_v, 1), sec=sec)
        expect_error(pc.source_var, (cell.soma(0.5)._ref_nai, 2), sec=sec)
        del sec
        expect_error(pc.target_var, (cell.ic, cell.vc._ref_amp3, 1))
        # source sid already in use
        expect_error(pc.source_var, (cell.soma(0.5)._ref_nai, 1), sec=cell.soma)

    # partrans update: could not find parameter index

    # pv2node checks the parent
    mkmodel(1)
    s1 = h.Section(name="dend")
    s2 = h.Section(name="soma")
    ic = h.IClamp(s1(0.5))
    pc.source_var(s1(0)._ref_v, rank, sec=s1)
    pc.target_var(ic, ic._ref_amp, rank)
    run()
    assert s1(0).v == ic.amp
    """
  # but following changes the source node and things get screwed up
  # because of continuing to use a freed Node*. The solution is
  # beyond the scope of this pull request and would involve replacing
  # description in terms of Node* with (Section*, arc_position)
  s1.connect(s2(.5))
  run()
  print(s1(0).v, ic.amp)
  assert(s1(0).v == ic.amp)
  """
    # non_vsrc_update property disappears from Node*
    s1.insert("pas")  # not allowed to uninsert ions :(
    pc.source_var(s1(0.5)._ref_e_pas, rank + 10, sec=s1)
    pc.target_var(ic, ic._ref_delay, rank + 10)
    run()
    assert s1(0.5).e_pas == ic.delay
    s1.uninsert("pas")
    expect_error(run, ())
    teardown()
    del ic, s1, s2

    # missing setup_transfer
    mkmodel(4)
    transfer1()
    expect_error(h.finitialize, (-65,))

    # round robin transfer v to ic.amp and vc.amp1, nai to vc.amp2
    ncell = 5
    mkmodel(ncell)
    transfer1()
    init_values()
    run()
    check_values()

    # nrnmpi_int_alltoallv_sparse
    h.nrn_sparse_partrans = 1
    mkmodel(5)
    transfer1()
    init_values()
    run()
    check_values()
    h.nrn_sparse_partrans = 0

    # impedance error (number of gap junction not equal to number of pc.transfer_var)
    imp = h.Impedance()
    if 0 in model[0]:
        imp.loc(model[0][0].soma(0.5))
    expect_error(imp.compute, (1, 1))
    del imp

    # For impedance, pc.target_var requires that its first arg be a reference to the POINT_PROCESS"
    mkmodel(2)
    if pc.gid_exists(0):
        cell = pc.gid2cell(0)
        pc.source_var(cell.soma(0.5)._ref_v, 1000, sec=cell.soma)
        cell.hgap[1] = h.HGap(cell.soma(0.5))
        pc.target_var(cell.hgap[1], cell.hgap[1]._ref_e, 1001)
    if pc.gid_exists(1):
        cell = pc.gid2cell(1)
        pc.source_var(cell.soma(0.5)._ref_v, 1001, sec=cell.soma)
        cell.hgap[0] = h.HGap(cell.soma(0.5))
        pc.target_var(cell.hgap[0], cell.hgap[0]._ref_e, 1000)
    pc.setup_transfer()
    imp = h.Impedance()
    h.finitialize(-65)
    if pc.gid_exists(0):
        imp.loc(pc.gid2cell(0).soma(0.5))
    expect_error(imp.compute, (10, 1, 100))
    del imp, cell

    # impedance
    ncell = 5
    mkmodel(ncell)
    mkgaps(list(range(ncell - 1)))
    pc.setup_transfer()
    imp = h.Impedance()
    h.finitialize(-65)
    if 0 in model[0]:
        imp.loc(model[0][0].soma(0.5))
    niter = imp.compute(10, 1, 100)
    if rank == 0:
        print("impedance iterations=%d" % niter)
    # tickle execution of target_ptr_update for one more line of coverage.
    if 0 in model[0]:
        model[0][0].hgap[1].loc(model[0][0].soma(0))
        model[0][0].hgap[1].loc(model[0][0].soma(0.5))
    niter = imp.compute(10, 1, 100)
    del imp

    # CoreNEURON gap file generation
    mkmodel(ncell)
    transfer1()

    # following is a bit tricky and need some user help in the docs.
    #  cannot be cache_efficient if general sparse matrix solver in effect.
    cvode = h.CVode()
    assert cvode.use_mxb(0) == 0
    assert cvode.cache_efficient(1) == 1

    pc.setup_transfer()
    h.finitialize(-65)
    pc.nrncore_write("tmp")

    # CoreNEURON: one thread empty of gaps
    mkmodel(1)
    transfer1()
    s = h.Section("dend")
    pc.set_gid2node(rank + 10, rank)
    pc.cell(rank + 10, h.NetCon(s(0.5)._ref_v, None, sec=s))
    pc.nthread(2)
    pc.setup_transfer()
    h.finitialize(-65)
    pc.nrncore_write("tmp")
    pc.nthread(1)
    teardown()
    del s

    # There are single thread circumstances where target POINT_PROCESS is needed
    s = h.Section("dend")
    pc.set_gid2node(rank, rank)
    pc.cell(rank, h.NetCon(s(0.5)._ref_v, None, sec=s))
    pc.source_var(s(0.5)._ref_v, rank, sec=s)
    ic = h.IClamp(s(0.5))
    pc.target_var(ic._ref_amp, rank)
    pc.setup_transfer()
    expect_error(h.finitialize, (-65,))
    teardown()
    del ic, s

    # threads
    mkmodel(ncell)
    transfer1()
    pc.nthread(2)
    init_values()
    run()
    check_values()
    pc.nthread(1)

    # extracellular means use v = vm+vext[0]
    for cell in model[0].values():
        cell.soma.insert("extracellular")
    init_values()
    run()
    check_values()

    teardown()


if __name__ == "__main__":
    test_partrans()
    pc.barrier()
    h.quit()



import numpy


def test_surface_area(neuron_instance):
    """Test that surface node all have surface area"""

    h, rxd, data, save_path = neuron_instance

    soma = h.Section(name="soma")
    soma.L = 1
    soma.diam = 1
    soma.nseg = 1

    rxd.set_solve_type(dimension=3)
    allsecs = soma.wholetree()
    r = rxd.Region(allsecs, dx=0.1)
    ca = rxd.Species(r)
    h.finitialize(-65)

    # true surface area
    sa = 1.5 * numpy.pi

    areas = numpy.array(ca.nodes.surface_area)

    # check surface area is 0.9 to 1.1 of the true value
    assert abs(1 - sum(areas) / sa) <= 0.1

    # check all surface nodes have non-zero surface area
    assert not any(areas[r._surface_nodes_by_seg[0]] == 0)



import neuron
import os
import numpy
import numpy.linalg

# from mayavi import mlab

# TODO: remove circular dependency
from . import surfaces


def _register_on_neighbor_map(the_map, pt, neighbor):
    # does not assume neighbor relations are bidirectional
    if pt in the_map:
        the_map[pt].append(neighbor)
    else:
        the_map[pt] = [neighbor]


class TriangularMesh:
    """
    A triangular mesh, typically of a surface.
    """

    def __init__(self, data):
        """
        Parameters
        ----------
        data : :class:`numpy.ndarray` or other iterable
            The raw data, listed in the form `(x0, y0, z0, x1, y1, z1, x2, y2, z2)` repeated.
        """
        self.data = data

    @property
    def x(self):
        """The x coordinates of the vertices."""
        return self.data[0::3]

    @property
    def y(self):
        """The y coordinates of the vertices."""
        return self.data[1::3]

    @property
    def z(self):
        """The z coordinates of the vertices."""
        return self.data[2::3]

    @property
    def faces(self):
        """A list of the triangles, described as lists of the indices of three points."""
        return [(i, i + 1, i + 2) for i in range(0, len(self.data) / 3, 3)]

    @property
    def area(self):
        """The sum of the areas of the constituent triangles."""
        return surfaces.tri_area(triangles)

    def has_unmatched_edge(self, precision=3):
        """Checks for edges that belong to only one triangle. True if they exist; else False.

        Parameters
        ----------
        precision : int, optional
            Number of digits after the decimal point to round to when comparing points.
        """

        scale_factor = 10**precision

        data = self.data
        pt_neighbor_map = {}
        for i in range(0, len(self.data), 9):
            pts = {}
            for j in range(3):
                pts[
                    tuple(
                        (scale_factor * data[i + 3 * j : i + 3 * j + 3]).round()
                        / scale_factor
                    )
                ] = 0
            pts = list(pts.keys())
            if len(pts) == 3:
                # only consider triangles of nonzero area
                for j in range(3):
                    for k in range(3):
                        if j != k:
                            _register_on_neighbor_map(pt_neighbor_map, pts[j], pts[k])
            # else:
            #    print '** discarded zero-area triangle **'
        edge_count = 0
        bad_pts = []
        # if no holes, each point should have each neighbor listed more than once
        for pt, neighbor_list in zip(
            list(pt_neighbor_map.keys()), list(pt_neighbor_map.values())
        ):
            count = {}
            for neighbor in neighbor_list:
                if neighbor not in count:
                    count[neighbor] = 1
                else:
                    count[neighbor] += 1
            for neighbor, ncount in zip(list(count.keys()), list(count.values())):
                if ncount <= 1:
                    if pt < neighbor:
                        # only print one direction of it
                        edge_count += 1
                        # mlab.plot3d([pt[0], neighbor[0]], [pt[1], neighbor[1]], [pt[2], neighbor[2]], color=(0,0,1))
                        if pt not in bad_pts:
                            bad_pts.append(pt)
                        if neighbor not in bad_pts:
                            bad_pts.append(neighbor)
                        # TODO: remove this; should never get here anyways
                        print(
                            (
                                "exposed edge: (%g, %g, %g) to (%g, %g, %g)"
                                % (pt + neighbor)
                            )
                        )

        if edge_count:
            return True
        # print 'total exposed edges: ', edge_count

        # bad_x = [pt[0] for pt in bad_pts]
        # bad_y = [pt[1] for pt in bad_pts]
        # bad_z = [pt[2] for pt in bad_pts]

        # mlab.points3d(bad_x, bad_y, bad_z, scale_factor=0.05, color=(0,0,1))

        return False



import neuron
import pytest


def test_hocbase():
    class MyList(neuron.HocBaseObject, hoc_type=neuron.h.Vector):
        pass

    assert issubclass(MyList, neuron.hoc.HocObject)
    assert issubclass(MyList, neuron.HocBaseObject)
    assert MyList._hoc_type == neuron.h.Vector


def test_hoc_template_hclass():
    neuron.h(
        """
            begintemplate A
            public x, s, o, xa, oa, f, p
            strdef s
            objref o, oa[2]
            double xa[3]
            proc init() { \
              x = $1 \
            }
            func f() { return $1*xa[$2] }
            proc p() { x += 1 }
            endtemplate A
        """
    )

    class A1(neuron.hclass(neuron.h.A)):
        def __new__(cls, arg):
            return super().__new__(cls, arg)

        def __init__(self, arg):
            self.bp = self.baseattr("p")

        def p(self):
            self.bp()
            return self.x

    a = A1(5)
    assert a.x == 5.0
    assert a.p() == 6.0
    b = A1(4)
    a.s = "one"
    b.s = "two"
    assert a.s == "one"
    assert b.s == "two"
    assert neuron.h.A[0].s == "one"
    assert a.p() == 7.0
    assert b.p() == 5.0
    a.a = 2
    b.a = 3
    assert a.a == 2
    assert b.a == 3
    assert neuron.h.List("A").count() == 2
    a = 1
    b = 1
    assert neuron.h.List("A").count() == 0


def test_pyobj_constructor():
    # Test that __new__ is required when __init__ is overridden
    with pytest.raises(TypeError):

        class PyObj(neuron.HocBaseObject, hoc_type=neuron.h.List):
            def __init__(self, first):
                super().__init__()
                self.append(first)

    class PyObj(neuron.HocBaseObject, hoc_type=neuron.h.List):
        def __new__(cls, first):
            return super().__new__(cls)

        def __init__(self, first):
            super().__init__()
            self.append(first)

    p = PyObj(neuron.h.List())
    assert p.count() == 1


def test_pyobj_def():
    class PyObj(neuron.HocBaseObject, hoc_type=neuron.h.List):
        def my_method(self, a):
            return a * 2

    p = PyObj()
    assert p.my_method(4) == 8


def test_pyobj_overloading():
    class PyObj(neuron.HocBaseObject, hoc_type=neuron.h.List):
        def append(self, i):
            self.last_appended = i
            return self.baseattr("append")(i)

    p = PyObj()
    p2 = PyObj()
    assert p.append(p) == 1
    assert p.count() == 1
    assert p[0] == p


def test_pyobj_inheritance():
    class PyObj(neuron.HocBaseObject, hoc_type=neuron.h.List):
        pass

    class MyObj(PyObj):
        pass

    with pytest.raises(TypeError):

        class MyObj2(PyObj):
            def __init__(self, arg):
                pass

    class List(neuron.HocBaseObject, hoc_type=neuron.h.List):
        def __new__(cls, *args, **kwargs):
            super().__new__(cls)

    class InitList(List):
        def __init__(self, *args):
            super().__init__()
            for arg in args:
                self.append(arg)

    l = InitList(neuron.h.List(), neuron.h.List())


def test_pyobj_composition():
    class A(neuron.HocBaseObject, hoc_type=neuron.h.List):
        pass

    class B(neuron.HocBaseObject, hoc_type=neuron.h.List):
        pass

    class C(neuron.HocBaseObject, hoc_type=neuron.h.Vector):
        pass

    with pytest.raises(TypeError):
        # Composition of different HOC types is impossible.
        class D(A, C):
            pass

    class E(A, B):
        pass

    assert E._hoc_type == neuron.h.List


class PickleTest(neuron.HocBaseObject, hoc_type=neuron.h.NetStim):
    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __init__(self, start, number, interval, noise):
        self.start = start
        self.number = number
        self.interval = interval
        self.noise = noise

    def __reduce__(self):
        return (
            self.__class__,
            (self.start, self.number, self.interval, self.noise),
        )


def test_pyobj_pickle():
    import pickle

    p = pickle.loads(pickle.dumps(PickleTest(10, 100, 1, 0)))
    assert p.__class__ is PickleTest
    assert p.start == 10



from neuron import h, rxd

h.load_file("stdrun.hoc")

from neuron.units import mM, ms

sim_time = 100

axon_terminal = h.Section(name="axon_terminal")
# synaptic_cleft = rxd.Region([axon_terminal], nrn_region='o')
axon_terminal_region = rxd.Region([axon_terminal], nrn_region="i")

ca_intra = rxd.Species(
    axon_terminal_region, name="ca", charge=2, d=0, initial=5e-4, atolscale=1e-6
)
F = rxd.Species(axon_terminal_region, name="F", initial=0.001)
FA = rxd.Species(axon_terminal_region, name="FA", initial=0)

kb = rxd.Parameter(axon_terminal_region, name="kb", value=1e16 / (mM**4 * ms))
ku = rxd.Parameter(axon_terminal_region, name="ku", value=0.1 / ms)


myreaction = rxd.Reaction(
    F + 4 * ca_intra, FA, kb, ku
)  # , regions=axon_terminal_region)

tvec = h.Vector().record(h._ref_t)
Fvec = h.Vector().record(F.nodes._ref_concentration)
FAvec = h.Vector().record(FA.nodes._ref_concentration)
h.finitialize(-65)
h.continuerun(10)

if __name__ == "__main__":
    from matplotlib import pyplot

    pyplot.plot(tvec, Fvec)
    pyplot.plot(tvec, FAvec)
    pyplot.show()



from neuron import h, crxd as rxd, gui
from neuron.crxd import initializer

h.CVode().active(1)

rxdsec = [h.Section(), h.Section()]
cyt = rxd.Region(rxdsec)
er = rxd.Region(rxdsec)
ca = rxd.Species([cyt, er])
caextrude = rxd.Rate(ca, 1)

# test for forced initialization before model fully described
initializer._do_init()
ip3 = rxd.Species(cyt)
h.tstop = 1000
h.run()



import pytest

from testutils import compare_data, tol


@pytest.fixture
def ecs_diffusion(neuron_instance):
    h, rxd, data, save_path = neuron_instance

    def make_model(nx, ny, nz, alpha, lambd, d=1.0, perm=None):
        dx = 10
        # the extracellular space
        ecs = rxd.Extracellular(
            -dx * nx / 2.0,
            -dx * ny / 2.0,
            -dx * nz / 2.0,
            dx * nx / 2.0,
            dx * ny / 2.0,
            dx * nz / 2.0,
            dx=dx,
            volume_fraction=alpha,
            tortuosity=lambd,
            permeability=perm,
        )

        # Who?
        k = rxd.Species(
            [ecs],
            name="k",
            d=d,
            charge=1,
            initial=lambda nd: 1
            if nd.x3d**2 + nd.y3d**2 + nd.z3d**2 < (2 * dx) ** 2
            else 0,
        )

        model = (ecs, k)
        return model

    yield (neuron_instance, make_model)


def test_ecs_diffusion_2d(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(11, 11, 1, 0.2, 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    k.d = 1
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_2d_cvode(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(11, 11, 1, 0.2, 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    k.d = 1
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_1d(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(1, 11, 1, 0.2, 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    k.nodes[0].d = 1.0
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_1d_cvode(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(1, 11, 1, 0.2, 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    k[ecs].d = (1, 1, 1)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_0d(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(1, 1, 1, 0.2, 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        assert k[ecs].states3d[0] == 1.0


def test_ecs_diffusion_0d_cvode(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    h, rxd, data, save_path = neuron_instance
    ecs, k = make_model(1, 1, 1, 0.2, 1.6, 0)
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        assert k[ecs].states3d[0] == 1.0


def test_ecs_diffusion_2d_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(11, 11, 1, 0.2, lambda x, y, z: 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    k[ecs].d = (1, 1, 1)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_2d_cvode_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(11, 11, 1, 0.2, lambda x, y, z: 1.6, 0)
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    k.nodes[0].d = (1, 1, 1)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_1d_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(1, 11, 1, 0.2, lambda x, y, z: 1.6)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_1d_cvode_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(1, 11, 1, 0.2, lambda x, y, z: 1.6)
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_0d_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(1, 1, 1, 0.2, lambda x, y, z: 1.6)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        assert k[ecs].states3d[0] == 1.0


def test_ecs_diffusion_0d_cvode_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    h, rxd, data, save_path = neuron_instance
    ecs, k = make_model(1, 1, 1, 0.2, lambda x, y, z: 1.6)
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        assert k[ecs].states3d[0] == 1.0


def test_ecs_diffusion_2d_alpha(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(11, 11, 1, lambda x, y, z: 0.2, 1.6)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_2d_cvode_alpha(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(11, 11, 1, lambda x, y, z: 0.2, 1.6)
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_1d_alpha(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(1, 11, 1, lambda x, y, z: 0.2, 1.6)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_1d_cvode_alpha(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(1, 11, 1, lambda x, y, z: 0.2, 1.6)
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_0d_alpha(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(1, 1, 1, lambda x, y, z: 0.2, 1.6)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    ecs, k = model
    if not save_path:
        assert k[ecs].states3d[0] == 1.0


def test_ecs_diffusion_0d_cvode_alpha(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    h, rxd, data, save_path = neuron_instance
    model = make_model(1, 1, 1, lambda x, y, z: 0.2, 1.6)
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    ecs, k = model
    if not save_path:
        assert k[ecs].states3d[0] == 1.0


def test_ecs_diffusion_anisotropic(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(11, 11, 11, 0.2, 1.6, d=(0.25, 0.5, 1.0))
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_cvode_anisotropic(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    model = make_model(11, 11, 11, 0.2, 1.6, d=(0.25, 0.5, 1.0))
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_3d_tort_state(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(
        11, 11, 11, 0.2, None, perm={"name": "permeability", "initial": 0}
    )
    h, rxd, data, save_path = neuron_instance
    for nd in ecs.permeability.nodes:
        nd.value = 100 / (10 + (nd.x3d + nd.y3d + nd.z3d) ** 2)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_3d_inhom_tort(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(
        11, 11, 11, 0.2, None, perm=lambda x, y, z: 100 / (10 + (x + y + z) ** 2)
    )
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_3d_inhom_tort_cvode(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(
        11, 11, 11, 0.2, None, perm=lambda x, y, z: 100 / (10 + (x + y + z) ** 2)
    )
    h, rxd, data, save_path = neuron_instance
    h.CVode().active(True)
    h.finitialize(1000)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_ecs_diffusion_3d_alpha_state(ecs_diffusion):
    neuron_instance, make_model = ecs_diffusion
    ecs, k = make_model(11, 11, 11, {"name": "alpha", "initial": 1}, 1.6)
    h, rxd, data, save_path = neuron_instance
    h.finitialize(1000)
    ecs.alpha.nodes.value = 0.2
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



"""test of pure diffusion

Robert A McDougal
March 2013 - June 2014
"""

from neuron import h, crxd as rxd
import numpy
import __main__

name = __main__.__file__
if name[-3:] == ".py":
    name = name[:-3]

h.load_file("stdrun.hoc")

dend = h.Section()
dend.diam = 2
dend.nseg = 101
dend.L = 100
rxd.set_solve_type(dimension=3)
diff_constant = 1
h.dt = h.dt * 50
r = rxd.Region(h.allsec(), dx=0.75)
ca = rxd.Species(
    r, d=diff_constant, initial=lambda node: 1 if 0.4 < node.x < 0.6 else 0
)

h.finitialize()
initial_amount = (
    numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)
).sum()

if __name__ == "__main__":
    from matplotlib import pyplot

    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)
        pyplot.plot([nd.x for nd in ca.nodes], ca.nodes.concentration)
    pyplot.tight_layout()
    pyplot.savefig("{0}.png".format(name))
    pyplot.show()
else:
    for t in [25, 50, 75, 100, 125]:
        h.continuerun(t)

final_amount = (
    numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)
).sum()

print(
    "initial {}\tfinal {}\tdiff {}".format(
        initial_amount, final_amount, initial_amount - final_amount
    )
)

pyplot.tight_layout()
pyplot.savefig("{0}.png".format(name))
pyplot.show()



def test_in(neuron_instance):
    """Test 1D diffusion in a single section"""

    h, rxd, data, save_path = neuron_instance

    soma = h.Section(name="soma")
    dend = h.Section(name="dend")
    soma.nseg = 3

    cyt = rxd.Region(h.allsec(), name="cyt")
    er = rxd.Region(h.allsec(), name="er")

    c = rxd.Species(cyt, name="c")

    for node in c.nodes(soma):
        assert node in soma
        assert node not in dend
        assert node in cyt
        assert node not in er

    for node in c.nodes(dend):
        assert node in dend
        assert node not in soma
        assert node in cyt
        assert node not in er



import pytest
import numpy
from math import exp
from testutils import compare_data, tol


@pytest.fixture
def ics_pure_diffusion_anisotropic(neuron_instance):
    """A model using intracellular diffusion in a single section in 3D"""

    def make_test(diff_constant):
        h, rxd, data, save_path = neuron_instance
        dend = h.Section(name="dend")
        dend.nseg = 11
        dend.pt3dclear()
        dend.pt3dadd(-3, 0, 0, 3)
        dend.pt3dadd(3, 0, 0, 3)
        rxd.set_solve_type(dimension=3)
        r = rxd.Region(h.allsec(), dx=0.75)
        ca = rxd.Species(
            r,
            d=diff_constant,
            initial=lambda nd: exp(
                -((nd.x3d - 0.375) ** 2 + (nd.y3d - 0.375) ** 2 + (nd.z3d - 0.375) ** 2)
            ),
        )
        return (dend, r, ca)

    yield (neuron_instance, make_test)


def test_pure_diffusion_3d_anisotropic_x(ics_pure_diffusion_anisotropic):
    """Test anisotropic without diffusion in the x direction with fixed step
    methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test([0, 0.1, 0.1])
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_y(ics_pure_diffusion_anisotropic):
    """Test anisotropic without diffusion in the y direction with fixed step
    methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test([0.1, 0, 0.1])
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_z(ics_pure_diffusion_anisotropic):
    """Test anisotropic without diffusion in the z direction with fixed step
    methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test([0.1, 0.1, 0])
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_x_cvode(ics_pure_diffusion_anisotropic):
    """Test anisotropic without diffusion in the x direction with variable step
    methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test([0, 0.1, 0.1])
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_y_cvode(ics_pure_diffusion_anisotropic):
    """Test anisotropic without diffusion in the y direction with variable step
    methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test([0.1, 0, 0.1])
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_z_cvode(ics_pure_diffusion_anisotropic):
    """Test anisotropic without diffusion in the z direction with variable step
    methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test([0.1, 0.1, 0])
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_x_inhom(ics_pure_diffusion_anisotropic):
    """Test inhomogeneous anisotropic diffusion without diffusion in the x
    direction with fixed step methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test(lambda nd, dr: 0 if dr == 0 else 0.1)
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_y_inhom(ics_pure_diffusion_anisotropic):
    """Test inhomogeneous anisotropic diffusion without diffusion in the y
    direction with fixed step methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test(lambda nd, dr: 0 if dr == 1 else 0.1)
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_z_inhom(ics_pure_diffusion_anisotropic):
    """Test inhomogeneous anisotropic diffusion without diffusion in the z
    direction with fixed step methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test(lambda nd, dr: 0 if dr == 2 else 0.1)
    h.dt *= 50
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_x_inhom_cvode(ics_pure_diffusion_anisotropic):
    """Test inhomogeneous anisotropic diffusion without diffusion in the x
    direction with variable step methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test(lambda nd, dr: 0 if dr == 0 else 0.1)
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_y_inhom_cvode(ics_pure_diffusion_anisotropic):
    """Test inhomogeneous anisotropic diffusion without diffusion in the y
    direction with variable step methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test(lambda nd, dr: 0 if dr == 1 else 0.1)
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol


def test_pure_diffusion_3d_anisotropic_z_inhom_cvode(ics_pure_diffusion_anisotropic):
    """Test inhomogeneous anisotropic diffusion without diffusion in the z
    direction with fixed step methods"""

    neuron_instance, make_test = ics_pure_diffusion_anisotropic
    h, rxd, data, save_path = neuron_instance
    dend, r, ca = make_test(lambda nd, dr: 0 if dr == 2 else 0.1)
    h.CVode().active(True)
    h.finitialize(-65)
    loss = -(numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    h.continuerun(125)
    loss += (numpy.array(ca.nodes.concentration) * numpy.array(ca.nodes.volume)).sum()
    if not save_path:
        assert loss < tol
    max_err = compare_data(data)
    if not save_path:
        assert max_err < tol
